#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations

import math
from typing import Iterable, no_type_check, Any
import copy

import PIL.Image
import numpy as np

from ezdxf.math import Vec2, BoundingBox2d
from ezdxf.colors import RGB
from ezdxf.path import Command
from ezdxf.version import __version__
from ezdxf.lldxf.validator import make_table_key as layer_key

from .type_hints import Color
from .backend import BackendInterface, BkPath2d, BkPoints2d, ImageData
from .config import Configuration, LineweightPolicy
from .properties import BackendProperties
from . import layout, recorder

is_pymupdf_installed = True
pymupdf: Any = None
try:
    import pymupdf  # type: ignore[import-untyped, no-redef]
except ImportError:
    print(
        "Python module PyMuPDF (AGPL!) is required: https://pypi.org/project/PyMuPDF/"
    )
    is_pymupdf_installed = False
# PyMuPDF docs: https://pymupdf.readthedocs.io/en/latest/

__all__ = ["PyMuPdfBackend", "is_pymupdf_installed"]

# PDF units are points (pt), 1 pt is 1/72 of an inch:
MM_TO_POINTS = 72.0 / 25.4  # 25.4 mm = 1 inch / 72
# psd does not work in PyMuPDF v1.22.3
SUPPORTED_IMAGE_FORMATS = ("png", "ppm", "pbm")


class PyMuPdfBackend(recorder.Recorder):
    """This backend uses the `PyMuPdf`_ package to create PDF, PNG, PPM and PBM output.
    This backend support content cropping at page margins.

    PyMuPDF is licensed under the `AGPL`_. Sorry, but it's the best package for the job
    I've found so far.

    Install package::

        pip install pymupdf

    .. _PyMuPdf: https://pypi.org/project/PyMuPDF/
    .. _AGPL: https://www.gnu.org/licenses/agpl-3.0.html

    """

    def __init__(self) -> None:
        super().__init__()
        self._init_flip_y = True

    def get_replay(
        self,
        page: layout.Page,
        *,
        settings: layout.Settings = layout.Settings(),
        render_box: BoundingBox2d | None = None,
    ) -> PyMuPdfRenderBackend:
        """Returns the PDF document as bytes.

        Args:
            page: page definition, see :class:`~ezdxf.addons.drawing.layout.Page`
            settings: layout settings, see :class:`~ezdxf.addons.drawing.layout.Settings`
            render_box: set explicit region to render, default is content bounding box
        """
        top_origin = True
        # This player changes the original recordings!
        player = self.player()
        if render_box is None:
            render_box = player.bbox()

        # the page origin (0, 0) is in the top-left corner.
        output_layout = layout.Layout(render_box, flip_y=self._init_flip_y)
        page = output_layout.get_final_page(page, settings)

        # DXF coordinates are mapped to PDF Units in the first quadrant
        settings = copy.copy(settings)
        settings.output_coordinate_space = get_coordinate_output_space(page)

        m = output_layout.get_placement_matrix(
            page, settings=settings, top_origin=top_origin
        )
        # transform content to the output coordinates space:
        player.transform(m)
        if settings.crop_at_margins:
            p1, p2 = page.get_margin_rect(top_origin=top_origin)  # in mm
            # scale factor to map page coordinates to output space coordinates:
            output_scale = settings.page_output_scale_factor(page)
            max_sagitta = 0.1 * MM_TO_POINTS  # curve approximation 0.1 mm
            # crop content inplace by the margin rect:
            player.crop_rect(p1 * output_scale, p2 * output_scale, max_sagitta)

        self._init_flip_y = False
        backend = self.make_backend(page, settings)
        player.replay(backend)
        return backend

    def get_pdf_bytes(
        self,
        page: layout.Page,
        *,
        settings: layout.Settings = layout.Settings(),
        render_box: BoundingBox2d | None = None,
    ) -> bytes:
        """Returns the PDF document as bytes.

        Args:
            page: page definition, see :class:`~ezdxf.addons.drawing.layout.Page`
            settings: layout settings, see :class:`~ezdxf.addons.drawing.layout.Settings`
            render_box: set explicit region to render, default is content bounding box
        """
        backend = self.get_replay(page, settings=settings, render_box=render_box)
        return backend.get_pdf_bytes()

    def get_pixmap_bytes(
        self,
        page: layout.Page,
        *,
        fmt="png",
        settings: layout.Settings = layout.Settings(),
        dpi: int = 96,
        alpha=False,
        render_box: BoundingBox2d | None = None,
    ) -> bytes:
        """Returns a pixel image as bytes, supported image formats:

        === =========================
        png Portable Network Graphics
        ppm Portable Pixmap (no alpha channel)
        pbm Portable Bitmap (no alpha channel)
        === =========================

        Args:
            page: page definition, see :class:`~ezdxf.addons.drawing.layout.Page`
            fmt: image format
            settings: layout settings, see :class:`~ezdxf.addons.drawing.layout.Settings`
            dpi: output resolution in dots per inch
            alpha: add alpha channel (transparency)
            render_box: set explicit region to render, default is content bounding box
        """
        if fmt not in SUPPORTED_IMAGE_FORMATS:
            raise ValueError(f"unsupported image format: '{fmt}'")
        backend = self.get_replay(page, settings=settings, render_box=render_box)
        try:
            pixmap = backend.get_pixmap(dpi=dpi, alpha=alpha)
            return pixmap.tobytes(output=fmt)
        except RuntimeError as e:
            print(f"PyMuPDF Runtime Error: {str(e)}")
            return b""

    @staticmethod
    def make_backend(
        page: layout.Page, settings: layout.Settings
    ) -> PyMuPdfRenderBackend:
        """Override this method to use a customized render backend."""
        return PyMuPdfRenderBackend(page, settings)


def get_coordinate_output_space(page: layout.Page) -> int:
    page_width_in_pt = int(page.width_in_mm * MM_TO_POINTS)
    page_height_in_pt = int(page.height_in_mm * MM_TO_POINTS)
    return max(page_width_in_pt, page_height_in_pt)


class PyMuPdfRenderBackend(BackendInterface):
    """Creates the PDF/PNG/PSD/SVG output.

    This backend requires some preliminary work, record the frontend output via the
    Recorder backend to accomplish the following requirements:

    - Move content in the first quadrant of the coordinate system.
    - The page is defined by the upper left corner in the origin (0, 0) and
      the lower right corner at (page-width, page-height)
    - The output coordinates are floats in 1/72 inch, scale the content appropriately
    - Replay the recorded output on this backend.

    .. important::

        Python module PyMuPDF is required: https://pypi.org/project/PyMuPDF/

    """

    def __init__(self, page: layout.Page, settings: layout.Settings) -> None:
        assert (
            is_pymupdf_installed
        ), "Python module PyMuPDF is required: https://pypi.org/project/PyMuPDF/"
        self.doc = pymupdf.open()
        self.doc.set_metadata(
            {
                "producer": f"PyMuPDF {pymupdf.version[0]}",
                "creator": f"ezdxf {__version__}",
            }
        )
        self.settings = settings
        self._optional_content_groups: dict[str, int] = {}
        self._stroke_width_cache: dict[float, float] = {}
        self._color_cache: dict[str, tuple[float, float, float]] = {}
        self.page_width_in_pt = int(page.width_in_mm * MM_TO_POINTS)
        self.page_height_in_pt = int(page.height_in_mm * MM_TO_POINTS)
        # LineweightPolicy.ABSOLUTE:
        self.min_lineweight = 0.05  # in mm, set by configure()
        self.lineweight_scaling = 1.0  # set by configure()
        self.lineweight_policy = LineweightPolicy.ABSOLUTE  # set by configure()

        # when the stroke width is too thin PDF viewers may get confused;
        self.abs_min_stroke_width = 0.1  # pt == 0.03528mm (arbitrary choice)

        # LineweightPolicy.RELATIVE:
        # max_stroke_width is determined as a certain percentage of settings.output_coordinate_space
        self.max_stroke_width: float = max(
            self.abs_min_stroke_width,
            int(settings.output_coordinate_space * settings.max_stroke_width),
        )
        # min_stroke_width is determined as a certain percentage of max_stroke_width
        self.min_stroke_width: float = max(
            self.abs_min_stroke_width,
            int(self.max_stroke_width * settings.min_stroke_width),
        )
        # LineweightPolicy.RELATIVE_FIXED:
        # all strokes have a fixed stroke-width as a certain percentage of max_stroke_width
        self.fixed_stroke_width: float = max(
            self.abs_min_stroke_width,
            int(self.max_stroke_width * settings.fixed_stroke_width),
        )
        self.page = self.doc.new_page(-1, self.page_width_in_pt, self.page_height_in_pt)
        # The page content is stored in a shared shape:
        self.content_shape = self.page.new_shape()
        # see also: https://github.com/pymupdf/PyMuPDF/issues/3800
        
    def get_pdf_bytes(self) -> bytes:
        return self.doc.tobytes()

    def get_pixmap(self, dpi: int, alpha=False):
        return self.page.get_pixmap(dpi=dpi, alpha=alpha)

    def get_svg_image(self) -> str:
        return self.page.get_svg_image()

    def set_background(self, color: Color) -> None:
        rgb = self.resolve_color(color)
        opacity = alpha_to_opacity(color[7:9])
        if color == (1.0, 1.0, 1.0) or opacity == 0.0:
            return
        shape = self.content_shape
        shape.draw_rect([0, 0, self.page_width_in_pt, self.page_height_in_pt])
        shape.finish(width=None, color=None, fill=rgb, fill_opacity=opacity)
        shape.commit()

    def get_optional_content_group(self, layer_name: str) -> int:
        if not self.settings.output_layers:
            return 0  # the default value of `oc` when not provided
        layer_name = layer_key(layer_name)
        if layer_name not in self._optional_content_groups:
            self._optional_content_groups[layer_name] = self.doc.add_ocg(
                name=layer_name,
                config=-1,
                on=True,
            )
        return self._optional_content_groups[layer_name]

    def finish_line(self, shape, properties: BackendProperties, close: bool) -> None:
        color = self.resolve_color(properties.color)
        width = self.resolve_stroke_width(properties.lineweight)
        shape.finish(
            width=width,
            color=color,
            fill=None,
            lineJoin=1,
            lineCap=1,
            stroke_opacity=alpha_to_opacity(properties.color[7:9]),
            closePath=close,
            oc=self.get_optional_content_group(properties.layer),
        )

    def finish_filling(self, shape, properties: BackendProperties) -> None:
        shape.finish(
            width=None,
            color=None,
            fill=self.resolve_color(properties.color),
            fill_opacity=alpha_to_opacity(properties.color[7:9]),
            lineJoin=1,
            lineCap=1,
            closePath=True,
            even_odd=True,
            oc=self.get_optional_content_group(properties.layer),
        )

    def resolve_color(self, color: Color) -> tuple[float, float, float]:
        key = color[:7]
        try:
            return self._color_cache[key]
        except KeyError:
            pass
        color_floats = RGB.from_hex(color).to_floats()
        self._color_cache[key] = color_floats
        return color_floats

    def resolve_stroke_width(self, width: float) -> float:
        try:
            return self._stroke_width_cache[width]
        except KeyError:
            pass
        stroke_width = self.min_stroke_width
        if self.lineweight_policy == LineweightPolicy.ABSOLUTE:
            stroke_width = (  # in points (pt) = 1/72 inch
                max(self.min_lineweight, width) * MM_TO_POINTS * self.lineweight_scaling
            )
        elif self.lineweight_policy == LineweightPolicy.RELATIVE:
            stroke_width = map_lineweight_to_stroke_width(
                width, self.min_stroke_width, self.max_stroke_width
            )
        stroke_width = max(self.abs_min_stroke_width, stroke_width)
        self._stroke_width_cache[width] = stroke_width
        return stroke_width

    def draw_point(self, pos: Vec2, properties: BackendProperties) -> None:
        shape = self.content_shape
        pos = Vec2(pos)
        shape.draw_line(pos, pos)
        self.finish_line(shape, properties, close=False)

    def draw_line(self, start: Vec2, end: Vec2, properties: BackendProperties) -> None:
        shape = self.content_shape
        shape.draw_line(Vec2(start), Vec2(end))
        self.finish_line(shape, properties, close=False)

    def draw_solid_lines(
        self, lines: Iterable[tuple[Vec2, Vec2]], properties: BackendProperties
    ) -> None:
        shape = self.content_shape
        for start, end in lines:
            shape.draw_line(start, end)
        self.finish_line(shape, properties, close=False)

    def draw_path(self, path: BkPath2d, properties: BackendProperties) -> None:
        if len(path) == 0:
            return
        shape = self.content_shape
        add_path_to_shape(shape, path, close=False)
        self.finish_line(shape, properties, close=False)

    def draw_filled_paths(
        self, paths: Iterable[BkPath2d], properties: BackendProperties
    ) -> None:
        shape = self.content_shape
        for p in paths:
            add_path_to_shape(shape, p, close=True)
        self.finish_filling(shape, properties)

    def draw_filled_polygon(
        self, points: BkPoints2d, properties: BackendProperties
    ) -> None:
        vertices = points.to_list()
        if len(vertices) < 3:
            return
        # pymupdf >= 1.23.19 does not accept Vec2() instances
        # input coordinates are page coordinates in pdf units
        shape = self.content_shape
        shape.draw_polyline(vertices)
        self.finish_filling(shape, properties)

    def draw_image(self, image_data: ImageData, properties: BackendProperties) -> None:
        transform = image_data.transform
        image = image_data.image
        height, width, depth = image.shape
        assert depth == 4

        corners = list(
            transform.transform_vertices(
                [Vec2(0, 0), Vec2(width, 0), Vec2(width, height), Vec2(0, height)]
            )
        )
        xs = [p.x for p in corners]
        ys = [p.y for p in corners]
        r = pymupdf.Rect((min(xs), min(ys)), (max(xs), max(ys)))

        # translation and non-uniform scale are handled by having the image stretch to fill the given rect.
        angle = (corners[1] - corners[0]).angle_deg
        need_rotate = not math.isclose(angle, 0.0)
        # already mirroring once to go from pixels (+y down) to wcs (+y up)
        # so a positive determinant means an additional reflection
        need_flip = transform.determinant() > 0

        if need_rotate or need_flip:
            pil_image = PIL.Image.fromarray(image, mode="RGBA")
            if need_flip:
                pil_image = pil_image.transpose(PIL.Image.Transpose.FLIP_TOP_BOTTOM)
            if need_rotate:
                pil_image = pil_image.rotate(
                    -angle,
                    resample=PIL.Image.Resampling.BICUBIC,
                    expand=True,
                    fillcolor=(0, 0, 0, 0),
                )
            image = np.asarray(pil_image)
            height, width, depth = image.shape

        pixmap = pymupdf.Pixmap(
            pymupdf.Colorspace(pymupdf.CS_RGB), width, height, bytes(image.data), True
        )
        # TODO: could improve by caching and re-using xrefs. If a document contains many
        #  identical images redundant copies will be stored for each one
        self.page.insert_image(
            r,
            keep_proportion=False,
            pixmap=pixmap,
            oc=self.get_optional_content_group(properties.layer),
        )

    def configure(self, config: Configuration) -> None:
        self.lineweight_policy = config.lineweight_policy
        if config.min_lineweight:
            # config.min_lineweight in 1/300 inch!
            min_lineweight_mm = config.min_lineweight * 25.4 / 300
            self.min_lineweight = max(0.05, min_lineweight_mm)
        self.lineweight_scaling = config.lineweight_scaling

    def clear(self) -> None:
        pass

    def finalize(self) -> None:
        self.content_shape.commit()

    def enter_entity(self, entity, properties) -> None:
        pass

    def exit_entity(self, entity) -> None:
        pass


@no_type_check
def add_path_to_shape(shape, path: BkPath2d, close: bool) -> None:
    start = path.start
    sub_path_start = start
    last_point = start
    for command in path.commands():
        end = command.end
        if command.type == Command.MOVE_TO:
            if close and not sub_path_start.isclose(end):
                shape.draw_line(start, sub_path_start)
            sub_path_start = end
        elif command.type == Command.LINE_TO:
            shape.draw_line(start, end)
        elif command.type == Command.CURVE3_TO:
            shape.draw_curve(start, command.ctrl, end)
        elif command.type == Command.CURVE4_TO:
            shape.draw_bezier(start, command.ctrl1, command.ctrl2, end)
        start = end
        last_point = end
    if close and not sub_path_start.isclose(last_point):
        shape.draw_line(last_point, sub_path_start)


def map_lineweight_to_stroke_width(
    lineweight: float,
    min_stroke_width: float,
    max_stroke_width: float,
    min_lineweight=0.05,  # defined by DXF
    max_lineweight=2.11,  # defined by DXF
) -> float:
    """Map the DXF lineweight in mm to stroke-width in viewBox coordinates."""
    lineweight = max(min(lineweight, max_lineweight), min_lineweight) - min_lineweight
    factor = (max_stroke_width - min_stroke_width) / (max_lineweight - min_lineweight)
    return min_stroke_width + round(lineweight * factor, 1)


def alpha_to_opacity(alpha: str) -> float:
    # stroke-opacity: 0.0 = transparent; 1.0 = opaque
    # alpha: "00" = transparent; "ff" = opaque
    if len(alpha):
        try:
            return int(alpha, 16) / 255
        except ValueError:
            pass
    return 1.0



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from typing import Iterable, cast
from pathlib import Path
import pytest
import ezdxf
from ezdxf import recover
from ezdxf.entities import MText

DATA = Path(__file__).parent / "data"
COLUMNS_R2007 = "mtext_columns_R2007.dxf"
COLUMNS_R2018 = "mtext_columns_R2018.dxf"


def load_mtext_entities(name: str) -> Iterable[MText]:
    doc = ezdxf.readfile(DATA / name)
    msp = doc.modelspace()
    entities = msp.query("MTEXT")
    return entities


def recover_mtext_entities(name: str) -> Iterable[MText]:
    doc, auditor = recover.readfile(DATA / name)
    msp = doc.modelspace()
    entities = msp.query("MTEXT")
    return entities


@pytest.fixture(scope="module", params=["load", "recover"])
def r2007(request):
    if request.param == "load":
        return load_mtext_entities(COLUMNS_R2007)
    elif request.param == "recover":
        return recover_mtext_entities(COLUMNS_R2007)


@pytest.fixture(scope="module", params=["load", "recover"])
def r2018(request):
    if request.param == "load":
        return load_mtext_entities(COLUMNS_R2018)
    elif request.param == "recover":
        return recover_mtext_entities(COLUMNS_R2018)


def test_load_mtext_columns_from_dxf_r2007(r2007):
    for mtext in r2007:
        assert mtext.has_columns is True
        columns = mtext.columns
        assert columns.count == 3
        assert len(columns.linked_columns) == 2
        db = mtext.doc.entitydb
        for column in columns.linked_columns:
            column = cast(MText, column)
            assert column.is_alive is True
            assert column.has_columns is False, "should not have columns"
            assert (
                column.dxf.handle in db
            ), "should be stored in the entity database"
            assert column.dxf.owner is None, "should not be stored in a layout"


def test_load_mtext_columns_from_dxf_r2018(r2018):
    for mtext in r2018:
        assert mtext.has_columns is True
        columns = mtext.columns
        assert columns.count == 3
        assert len(columns.linked_columns) == 0


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2010-2021, Manfred Moitzi
# License: MIT License
from ezdxf.math import Vec3, EulerSpiral
from math import isclose

expected_points = [
    (0.0, 0.0),
    (0.4999511740825297, 0.0052079700401204106),
    (0.99843862987320509, 0.041620186803547267),
    (1.4881781381789292, 0.13983245006538086),
    (1.9505753764262783, 0.32742809475246343),
    (2.3516635639763064, 0.62320387651494735),
    (2.6419212729287223, 1.0273042715153904),
    (2.7637635905799862, 1.5086926753401932),
    (2.6704397998515645, 1.9952561538526452),
    (2.3566156629790327, 2.3766072153088964),
    (1.8936094203448928, 2.5322897289776636),
]


def test_approximate():
    spiral = EulerSpiral(2.0)
    results = spiral.approximate(5, 10)
    for expected, result in zip(expected_points, results):
        assert Vec3(expected).isclose(result)


def test_radius():
    spiral = EulerSpiral(2.0)
    assert isclose(spiral.radius(1), 4.0)
    assert isclose(spiral.radius(0), 0.0)


def test_tangent():
    spiral = EulerSpiral(2.0)
    assert isclose(spiral.tangent(1).angle, 0.125)


def test_distance():
    spiral = EulerSpiral(2.0)
    assert isclose(spiral.distance(10), 0.4)


def test_circle_midpoint():
    spiral = EulerSpiral(2.0)
    m = spiral.circle_center(2.0)
    assert m.isclose((0.9917242992178723, 2.082593218533209, 0))


def test_as_bspline():
    spiral = EulerSpiral(2.0)
    spline = spiral.bspline(5, 10)
    assert spline.degree == 3
    assert spline.max_t == 5
    results = spline.approximate(10)
    for expected, result in zip(expected_points, results):
        assert Vec3(expected).isclose(result)



#  Copyright (c) 2020-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from typing_extensions import Self
from ezdxf.lldxf import validator
from ezdxf.lldxf.const import SUBCLASS_MARKER
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.math import Vec3, Vec2, NULLVEC, X_AXIS, Y_AXIS
from .dxfentity import base_class, SubclassProcessor, DXFEntity
from .dxfobj import DXFObject
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.entities.dxfns import DXFNamespace
    from ezdxf import xref

__all__ = ["PlotSettings", "DXFLayout"]

acdb_plot_settings = DefSubclass(
    "AcDbPlotSettings",
    {
        # acdb_plot_settings is also part of LAYOUT and LAYOUT has a 'name' attribute
        "page_setup_name": DXFAttr(1, default=""),

        # An optional empty string selects the default printer/plotter
        "plot_configuration_file": DXFAttr(2, default="", optional=True),
        "paper_size": DXFAttr(4, default="A3"),
        "plot_view_name": DXFAttr(6, default=""),
        "left_margin": DXFAttr(40, default=7.5),  # in mm
        "bottom_margin": DXFAttr(41, default=20),  # in mm
        "right_margin": DXFAttr(42, default=7.5),  # in mm
        "top_margin": DXFAttr(43, default=20),  # in mm
        "paper_width": DXFAttr(44, default=420),  # in mm
        "paper_height": DXFAttr(45, default=297),  # in mm
        "plot_origin_x_offset": DXFAttr(46, default=0.0),  # in mm
        "plot_origin_y_offset": DXFAttr(47, default=0.0),  # in mm
        "plot_window_x1": DXFAttr(48, default=0.0),
        "plot_window_y1": DXFAttr(49, default=0.0),
        "plot_window_x2": DXFAttr(140, default=0.0),
        "plot_window_y2": DXFAttr(141, default=0.0),
        # Numerator of custom print scale: real world (paper) units:
        "scale_numerator": DXFAttr(142, default=1.0),
        # Denominator of custom print scale: drawing units:
        "scale_denominator": DXFAttr(143, default=1.0),
        # Plot layout flags:
        # 1 = plot viewport borders
        # 2 = show plot-styles
        # 4 = plot centered
        # 8 = plot hidden == hide paperspace entities?
        # 16 = use standard scale
        # 32 = plot with plot-styles
        # 64 = scale lineweights
        # 128 = plot entity lineweights
        # 512 = draw viewports first
        # 1024 = model type
        # 2048 = update paper
        # 4096 = zoom to paper on update
        # 8192 = initializing
        # 16384 = prev plot-init
        # the "Plot transparencies" option is stored in the XDATA section
        "plot_layout_flags": DXFAttr(70, default=688),
        # Plot paper units:
        # 0 = Plot in inches
        # 1 = Plot in millimeters
        # 2 = Plot in pixels
        "plot_paper_units": DXFAttr(
            72,
            default=1,
            validator=validator.is_in_integer_range(0, 3),
            fixer=RETURN_DEFAULT,
        ),
        # Plot rotation:
        # 0 = No rotation
        # 1 = 90 degrees counterclockwise
        # 2 = Upside-down
        # 3 = 90 degrees clockwise
        "plot_rotation": DXFAttr(
            73,
            default=0,
            validator=validator.is_in_integer_range(0, 4),
            fixer=RETURN_DEFAULT,
        ),
        # Plot type:
        # 0 = Last screen display
        # 1 = Drawing extents
        # 2 = Drawing limits
        # 3 = View specified by code 6
        # 4 = Window specified by codes 48, 49, 140, and 141
        # 5 = Layout information
        "plot_type": DXFAttr(
            74,
            default=5,
            validator=validator.is_in_integer_range(0, 6),
            fixer=RETURN_DEFAULT,
        ),
        # Associated CTB-file
        "current_style_sheet": DXFAttr(7, default=""),
        # Standard scale type:
        # 0 = Scaled to Fit
        # 1 = 1/128"=1'
        # 2 = 1/64"=1'
        # 3 = 1/32"=1'
        # 4 = 1/16"=1'
        # 5 = 3/32"=1'
        # 6 = 1/8"=1'
        # 7 = 3/16"=1'
        # 8 = 1/4"=1'
        # 9 = 3/8"=1'
        # 10 = 1/2"=1'
        # 11 = 3/4"=1'
        # 12 = 1"=1'
        # 13 = 3"=1'
        # 14 = 6"=1'
        # 15 = 1'=1'
        # 16 = 1:1
        # 17 = 1:2
        # 18 = 1:4
        # 19 = 1:8
        # 20 = 1:10
        # 21 = 1:16
        # 22 = 1:20
        # 23 = 1:30
        # 24 = 1:40
        # 25 = 1:50
        # 26 = 1:100
        # 27 = 2:1
        # 28 = 4:1
        # 29 = 8:1
        # 30 = 10:1
        # 31 = 100:1
        # 32 = 1000:1
        "standard_scale_type": DXFAttr(
            75,
            default=16,
            validator=validator.is_in_integer_range(0, 33),
            fixer=RETURN_DEFAULT,
        ),
        # Shade plot mode:
        # 0 = As Displayed
        # 1 = Wireframe
        # 2 = Hidden
        # 3 = Rendered
        "shade_plot_mode": DXFAttr(
            76,
            default=0,
            validator=validator.is_in_integer_range(0, 4),
            fixer=RETURN_DEFAULT,
        ),
        # Shade plot resolution level:
        # 0 = Draft
        # 1 = Preview
        # 2 = Normal
        # 3 = Presentation
        # 4 = Maximum
        # 5 = Custom
        "shade_plot_resolution_level": DXFAttr(
            77,
            default=2,
            validator=validator.is_in_integer_range(0, 6),
            fixer=RETURN_DEFAULT,
        ),
        # Valid range: 100 to 32767, Only applied when the shade_plot_resolution
        # level is set to 5 (Custom)
        "shade_plot_custom_dpi": DXFAttr(
            78,
            default=300,
            validator=validator.is_in_integer_range(100, 32768),
            fixer=validator.fit_into_integer_range(100, 32768),
        ),
        # Factor for unit conversion (mm -> inches)
        # 147: DXF Reference error: 'A floating point scale factor that represents
        # the standard scale value specified in code 75'
        "unit_factor": DXFAttr(
            147,
            default=1.0,
            validator=validator.is_greater_zero,
            fixer=RETURN_DEFAULT,
        ),
        "paper_image_origin_x": DXFAttr(148, default=0),
        "paper_image_origin_y": DXFAttr(149, default=0),
        "shade_plot_handle": DXFAttr(333, optional=True),
    },
)
acdb_plot_settings_group_codes = group_code_mapping(acdb_plot_settings)

# The "Plot transparencies" option is stored in the XDATA section of the
# LAYOUT entity:
# 1001
# PLOTTRANSPARENCY
# 1071
# 1


@register_entity
class PlotSettings(DXFObject):
    DXFTYPE = "PLOTSETTINGS"
    DXFATTRIBS = DXFAttributes(base_class, acdb_plot_settings)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(dxf, acdb_plot_settings_group_codes, 1)
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_plot_settings.name)

        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "page_setup_name",
                "plot_configuration_file",
                "paper_size",
                "plot_view_name",
                "left_margin",
                "bottom_margin",
                "right_margin",
                "top_margin",
                "paper_width",
                "paper_height",
                "plot_origin_x_offset",
                "plot_origin_y_offset",
                "plot_window_x1",
                "plot_window_y1",
                "plot_window_x2",
                "plot_window_y2",
                "scale_numerator",
                "scale_denominator",
                "plot_layout_flags",
                "plot_paper_units",
                "plot_rotation",
                "plot_type",
                "current_style_sheet",
                "standard_scale_type",
                "shade_plot_mode",
                "shade_plot_resolution_level",
                "shade_plot_custom_dpi",
                "unit_factor",
                "paper_image_origin_x",
                "paper_image_origin_y",
            ],
        )

    def register_resources(self, registry: xref.Registry) -> None:
        super().register_resources(registry)
        registry.add_handle(self.dxf.get("shade_plot_handle"))

    def map_resources(self, clone: Self, mapping: xref.ResourceMapper) -> None:
        super().map_resources(clone, mapping)
        shade_plot_handle = self.dxf.get("shade_plot_handle")
        if shade_plot_handle and shade_plot_handle != "0":
            clone.dxf.shade_plot_handle = mapping.get_handle(shade_plot_handle)
        else:
            clone.dxf.discard("shade_plot_handle")


acdb_layout = DefSubclass(
    "AcDbLayout",
    {
        # Layout name:
        "name": DXFAttr(1, default="Layoutname"),
        # Flag (bit-coded) to control the following:
        # 1 = Indicates the PSLTSCALE value for this layout when this layout is current
        # 2 = Indicates the LIMCHECK value for this layout when this layout is current
        "layout_flags": DXFAttr(70, default=1),
        # Tab order: This number is an ordinal indicating this layout's ordering in
        # the tab control that is attached to the AutoCAD drawing frame window.
        # Note that the "Model" tab always appears as the first tab regardless of
        # its tab order.
        "taborder": DXFAttr(71, default=1),
        # Minimum limits:
        "limmin": DXFAttr(10, xtype=XType.point2d, default=Vec2(0, 0)),
        # Maximum limits:
        "limmax": DXFAttr(11, xtype=XType.point2d, default=Vec2(420, 297)),
        # Insertion base point for this layout:
        "insert_base": DXFAttr(12, xtype=XType.point3d, default=NULLVEC),
        # Minimum extents for this layout:
        "extmin": DXFAttr(14, xtype=XType.point3d, default=Vec3(1e20, 1e20, 1e20)),
        # Maximum extents for this layout:
        "extmax": DXFAttr(15, xtype=XType.point3d, default=Vec3(-1e20, -1e20, -1e20)),
        "elevation": DXFAttr(146, default=0.0),
        "ucs_origin": DXFAttr(13, xtype=XType.point3d, default=NULLVEC),
        "ucs_xaxis": DXFAttr(
            16,
            xtype=XType.point3d,
            default=X_AXIS,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        "ucs_yaxis": DXFAttr(
            17,
            xtype=XType.point3d,
            default=Y_AXIS,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        # Orthographic type of UCS:
        # 0 = UCS is not orthographic
        # 1 = Top
        # 2 = Bottom
        # 3 = Front
        # 4 = Back
        # 5 = Left
        # 6 = Right
        "ucs_type": DXFAttr(
            76,
            default=1,
            validator=validator.is_in_integer_range(0, 7),
            fixer=RETURN_DEFAULT,
        ),
        # Handle of parent BLOCK_RECORD
        "block_record_handle": DXFAttr(330),
        # Handle to the viewport that was last active in this
        # layout when the layout was current:
        "viewport_handle": DXFAttr(331),
        # Handle of AcDbUCSTableRecord if UCS is a named
        # UCS. If not present, then UCS is unnamed
        "ucs_handle": DXFAttr(345),
        # Handle of AcDbUCSTableRecord of base UCS if UCS is
        # orthographic (76 code is non-zero). If not present and
        # 76 code is non-zero, then base UCS is taken to be WORLD
        "base_ucs_handle": DXFAttr(346),
    },
)
acdb_layout_group_codes = group_code_mapping(acdb_layout)


@register_entity
class DXFLayout(PlotSettings):
    DXFTYPE = "LAYOUT"
    DXFATTRIBS = DXFAttributes(base_class, acdb_plot_settings, acdb_layout)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(dxf, acdb_layout_group_codes, 2)
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        # Set correct model type flag
        self.set_flag_state(1024, self.dxf.name.upper() == "MODEL", "plot_layout_flags")
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_layout.name)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "name",
                "layout_flags",
                "taborder",
                "limmin",
                "limmax",
                "insert_base",
                "extmin",
                "extmax",
                "elevation",
                "ucs_origin",
                "ucs_xaxis",
                "ucs_yaxis",
                "ucs_type",
                "block_record_handle",
                "viewport_handle",
                "ucs_handle",
                "base_ucs_handle",
            ],
        )

    def register_resources(self, registry: xref.Registry) -> None:
        super().register_resources(registry)
        registry.add_handle(self.dxf.get("ucs_handle"))
        registry.add_handle(self.dxf.get("base_ucs_handle"))

    def map_resources(self, clone: Self, mapping: xref.ResourceMapper) -> None:
        super().map_resources(clone, mapping)

        # The content of paperspace layouts is not copied automatically and the
        # associated BLOCK_RECORD is created and assigned in a special method.
        mapping.map_existing_handle(self, clone, "ucs_handle", optional=True)
        mapping.map_existing_handle(self, clone, "base_ucs_handle", optional=True)
        mapping.map_existing_handle(self, clone, "viewport_handle", optional=True)



# Copyright (c) 2018-2019 Manfred Moitzi
# License: MIT License
import ezdxf
from ezdxf.entities.dxfobj import VBAProject
from ezdxf.lldxf.tagwriter import TagCollector


def test_vba_project():
    doc = ezdxf.new("R2007")
    vba = doc.objects.new_entity("VBA_PROJECT", {})  # type: VBAProject
    assert vba.dxftype() == "VBA_PROJECT"
    assert len(vba.data) == 0

    data = (
        b"abcdefghij" * 100
    )  # 1000 bytes => 8 DXFBinaryTags() with <= 127 bytes
    assert len(data) == 1000

    vba.data = data
    collector = TagCollector()
    vba.export_data(collector)
    tags = collector.tags
    assert len(tags) == 8
    # 127 bytes encoded as 254 char string
    assert len(tags[-2].value) == 127
    assert len(tags[-1].value) == 1000 - (127 * 7)

    assert vba.data == data


VBA_PROJECT = """0
VBA_PROJECT
5
0
330
0
100
AcDbVbaProject
90
6
310
A0B0C0D0E0F0
"""


def test_load_vba_project():
    vba = VBAProject.from_text(VBA_PROJECT)
    assert len(vba.data) == 6
    assert vba.data == b"\xA0\xB0\xC0\xD0\xE0\xF0"



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
#  module to set application specific data
from __future__ import annotations
from typing import TYPE_CHECKING

from ezdxf.lldxf import const, validator
from ezdxf.math import BoundingBox
from ezdxf import bbox
from ezdxf.enums import EndCaps, JoinStyle

if TYPE_CHECKING:
    from ezdxf.document import Drawing

CURRENT_LAYER = "$CLAYER"
CURRENT_COLOR = "$CECOLOR"
CURRENT_LINETYPE = "$CELTYPE"
CURRENT_LINEWEIGHT = "$CELWEIGHT"
CURRENT_LINETYPE_SCALE = "$CELTSCALE"
CURRENT_TEXTSTYLE = "$TEXTSTYLE"
CURRENT_DIMSTYLE = "$DIMSTYLE"
EXTMIN = "$EXTMIN"
EXTMAX = "$EXTMAX"


def set_current_layer(doc: Drawing, name: str):
    """Set current layer."""
    if name not in doc.layers:
        raise const.DXFValueError(f'undefined layer: "{name}"')
    doc.header[CURRENT_LAYER] = name


def set_current_color(doc: Drawing, color: int):
    """Set current :ref:`ACI`."""
    if not validator.is_valid_aci_color(color):
        raise const.DXFValueError(f'invalid ACI color value: "{color}"')
    doc.header[CURRENT_COLOR] = color


def set_current_linetype(doc: Drawing, name: str):
    """Set current linetype."""
    if name not in doc.linetypes:
        raise const.DXFValueError(f'undefined linetype: "{name}"')
    doc.header[CURRENT_LINETYPE] = name


def set_current_lineweight(doc: Drawing, lineweight: int):
    """Set current lineweight, see :ref:`lineweights` reference for valid
    values.
    """
    if not validator.is_valid_lineweight(lineweight):
        raise const.DXFValueError(f'invalid lineweight value: "{lineweight}"')
    doc.header[CURRENT_LINEWEIGHT] = lineweight


def set_current_linetype_scale(doc: Drawing, scale: float):
    """Set current linetype scale."""
    if scale <= 0.0:
        raise const.DXFValueError(f'invalid linetype scale: "{scale}"')
    doc.header[CURRENT_LINETYPE_SCALE] = scale


def set_current_textstyle(doc: Drawing, name: str):
    """Set current text style."""
    if name not in doc.styles:
        raise const.DXFValueError(f'undefined textstyle: "{name}"')
    doc.header[CURRENT_TEXTSTYLE] = name


def set_current_dimstyle(doc: Drawing, name: str):
    """Set current dimstyle."""
    if name not in doc.dimstyles:
        raise const.DXFValueError(f'undefined dimstyle: "{name}"')
    doc.header[CURRENT_DIMSTYLE] = name


def set_current_dimstyle_attribs(doc: Drawing, name: str):
    """Set current dimstyle and copy all dimstyle attributes to the HEADER section."""
    set_current_dimstyle(doc, name)
    dimstyle = doc.dimstyles.get(name)
    dimstyle.copy_to_header(doc)


def restore_wcs(doc: Drawing):
    """Restore the UCS settings in the HEADER section to the :ref:`WCS` and
    reset all active viewports to the WCS.
    """
    doc.header.reset_wcs()
    for vport in doc.viewports.get_config("*Active"):
        vport.reset_wcs()


def update_extents(doc: Drawing) -> BoundingBox:
    """Calculate the extents of the model space, update the HEADER variables
    $EXTMIN and $EXTMAX and returns the result as :class:`ezdxf.math.BoundingBox`.
    Note that this function uses the :mod:`ezdxf.bbox` module to calculate the
    extent of the model space. This module is not very fast and not very
    accurate for text and ignores all :term:`ACIS` based entities.

    The function updates only the values in the HEADER section, to zoom the
    active viewport to this extents, use this recipe::

        import ezdxf
        from ezdxf import zoom, appsettings

        doc = ezdxf.readfile("your.dxf")
        extents = appsettings.update_extents(doc)
        zoom.center(doc.modelspace(), extents.center, extents.size)

    .. seealso::

        - the :mod:`ezdxf.bbox` module to understand the limitations of the
          extent calculation
        - the :mod:`ezdxf.zoom` module

    """
    msp = doc.modelspace()
    extents = bbox.extents(msp, fast=True)
    if extents.has_data:
        msp.dxf.extmin = extents.extmin
        msp.dxf.extmax = extents.extmax
        doc.header[EXTMIN] = extents.extmin
        doc.header[EXTMAX] = extents.extmax
    return extents


def show_lineweight(doc: Drawing, state=True) -> None:
    """The CAD application or DXF viewer should show lines and curves with
    "thickness" (lineweight) if `state` is ``True``.
    """
    doc.header["$LWDISPLAY"] = int(state)


def set_lineweight_display_style(
    doc: Drawing, end_caps: EndCaps, join_style: JoinStyle
) -> None:
    """Set the style of end caps and joints for linear entities when displaying
    line weights. These settings only affect objects created afterwards.
    """
    doc.header["$ENDCAPS"] = int(end_caps)
    doc.header["$JOINSTYLE"] = int(join_style)



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
import ezdxf
from ezdxf.render.arrows import ARROWS
from ezdxf.layouts import VirtualLayout


def test_filled_solid_arrow():
    # special name: no name ""
    assert "" in ARROWS
    ARROWS.is_acad_arrow("")


def test_arrow_name():
    assert ARROWS.arrow_name("_CLOSEDFILLED") == ""
    assert ARROWS.arrow_name("") == ""
    assert ARROWS.arrow_name("_DOTSMALL") == "DOTSMALL"
    assert ARROWS.arrow_name("_boxBlank") == "BOXBLANK"
    assert ARROWS.arrow_name("EZ_ARROW") == "EZ_ARROW"
    assert ARROWS.arrow_name("abcdef") == "abcdef"


def test_closed_arrow_doc_r12():
    doc = ezdxf.new(dxfversion="R12", setup=True)
    blocks = doc.blocks
    name = ARROWS.create_block(blocks, ARROWS.closed)
    arrow_entities = list(blocks.get(name))
    assert arrow_entities[0].dxftype() == "POLYLINE"


def test_closed_arrow_doc_r2000():
    doc = ezdxf.new(dxfversion="R2000", setup=True)
    blocks = doc.blocks
    name = ARROWS.create_block(blocks, ARROWS.closed)
    arrow_entities = list(blocks.get(name))
    assert arrow_entities[0].dxftype() == "LWPOLYLINE"


def test_render_arrow():
    layout = VirtualLayout()
    ARROWS.render_arrow(layout, ARROWS.closed, insert=(0, 0, 0))
    assert len(layout) > 0


def test_virtual_entities():
    entities = list(ARROWS.virtual_entities(ARROWS.closed, insert=(0, 0, 0)))
    assert len(entities) > 0



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.math import Vec3, BoundingBox
from ezdxf import disassemble
from ezdxf.entities import factory


def test_do_nothing():
    assert list(disassemble.recursive_decompose([])) == []
    assert list(disassemble.to_primitives([])) == []
    assert list(disassemble.to_vertices([])) == []


def test_convert_unsupported_entity_to_primitive():
    p = disassemble.make_primitive(factory.new("3DSOLID"))
    assert p.path is None
    assert p.mesh is None
    assert p.is_empty is True
    assert list(p.vertices()) == []
    assert p.bbox().has_data is False


def test_multiple_unsupported_entities_to_vertices():
    w = factory.new("3DSOLID")
    primitives = list(disassemble.to_primitives([w, w, w]))
    assert len(primitives) == 3, "3 empty primitives expected"
    vertices = list(disassemble.to_vertices(primitives))
    assert len(vertices) == 0, "no vertices expected"


def test_point_to_primitive():
    e = factory.new("POINT", dxfattribs={"location": (1, 2, 3)})
    p = disassemble.make_primitive(e)
    assert p.is_empty is False
    assert p.path is not None
    assert p.mesh is None
    assert list(p.vertices()) == [(1, 2, 3)]
    assert p.bbox().extmin == (1, 2, 3)
    assert p.bbox().extmax == (1, 2, 3)


def test_line_to_primitive():
    start = Vec3(1, 2, 3)
    end = Vec3(4, 5, 6)
    e = factory.new("LINE", dxfattribs={"start": start, "end": end})
    p = disassemble.make_primitive(e)
    assert p.is_empty is False
    assert p.path is not None
    assert p.mesh is None
    assert list(p.vertices()) == [start, end]


def test_lwpolyline_to_primitive():
    p1 = Vec3(1, 1)
    p2 = Vec3(2, 2)
    p3 = Vec3(3, 3)
    e = factory.new("LWPOLYLINE")
    e.append_points([p1, p2, p3], format="xy")
    p = disassemble.make_primitive(e)
    assert p.is_empty is False
    assert p.path is not None
    assert p.mesh is None
    assert list(p.vertices()) == [p1, p2, p3]
    assert p.bbox().extmin == p1
    assert p.bbox().extmax == p3


def test_circle_to_primitive():
    e = factory.new("CIRCLE", dxfattribs={"radius": 5})
    p = disassemble.make_primitive(e)
    assert p.is_empty is False
    assert p.path is not None
    assert p.mesh is None
    assert len(list(p.vertices())) > 32
    box = p.bbox()
    assert box.extmin == (-5, -5, 0)
    assert box.extmax == (5, 5, 0)


def test_arc_to_primitive():
    e = factory.new("ARC", dxfattribs={"radius": 5})
    p = disassemble.make_primitive(e)
    assert p.is_empty is False
    assert p.path is not None
    assert p.mesh is None
    assert len(list(p.vertices())) > 32
    box = p.bbox()
    assert box.extmin == (-5, -5, 0)
    assert box.extmax == (5, 5, 0)


def test_ellipse_to_primitive():
    e = factory.new("ELLIPSE", dxfattribs={"major_axis": (5, 0)})
    p = disassemble.make_primitive(e)
    assert p.is_empty is False
    assert p.path is not None
    assert p.mesh is None
    assert len(list(p.vertices())) > 32
    box = p.bbox()
    assert box.extmin == (-5, -5, 0)
    assert box.extmax == (5, 5, 0)


def test_spline_to_primitive():
    e = factory.new("SPLINE")
    e.control_points = [(0, 0), (3, 2), (6, -2), (9, 4)]
    p = disassemble.make_primitive(e)
    assert p.is_empty is False
    assert p.path is not None
    assert p.mesh is None
    assert len(list(p.vertices())) > 20
    assert len(list(p.path.flattening(0.01))) > 20
    # fast bounding box calculation
    box = p.bbox(fast=True)
    assert box.extmin == (0, -2, 0)
    assert box.extmax == (9, 4, 0)

    # precise bounding box calculation
    box = p.bbox(fast=False)
    assert box.extmin == (0, 0, 0)  # correct: checked visually
    assert box.extmax == (9, 4, 0)


def test_mesh_entity_to_primitive():
    from ezdxf.layouts import VirtualLayout
    from ezdxf.render.forms import cube

    vl = VirtualLayout()
    mesh_entity = cube().render_mesh(vl)
    assert mesh_entity.dxftype() == "MESH"

    p = disassemble.make_primitive(mesh_entity)
    assert p.is_empty is False
    assert p.path is None
    mesh_builder = p.mesh
    assert mesh_builder is not None
    assert p.is_empty is False

    assert len(mesh_builder.vertices) == 8
    assert len(mesh_builder.faces) == 6
    assert len(list(p.vertices())) == 8

    box = p.bbox()
    assert box.extmin == (-0.5, -0.5, -0.5)
    assert box.extmax == (0.5, 0.5, 0.5)


@pytest.mark.parametrize("dxftype", ["SOLID", "TRACE", "3DFACE"])
def test_from_quadrilateral_with_3_points(dxftype):
    entity = factory.new(dxftype)
    entity.dxf.vtx0 = (0, 0, 0)
    entity.dxf.vtx1 = (1, 0, 0)
    entity.dxf.vtx2 = (1, 1, 0)
    entity.dxf.vtx3 = (1, 1, 0)  # last point is repeated
    p = disassemble.make_primitive(entity)
    assert p.is_empty is False
    assert p.path is not None
    assert p.mesh is None
    assert len(list(p.vertices())) == 4, "expected closed path"

    box = p.bbox()
    assert box.extmin == (0, 0)
    assert box.extmax == (1, 1)


@pytest.mark.parametrize("dxftype", ["SOLID", "TRACE", "3DFACE"])
def test_from_quadrilateral_with_4_points(dxftype):
    entity = factory.new(dxftype)
    entity.dxf.vtx0 = (0, 0, 0)
    entity.dxf.vtx1 = (1, 0, 0)
    entity.dxf.vtx2 = (1, 1, 0)
    entity.dxf.vtx3 = (0, 1, 0)
    p = disassemble.make_primitive(entity)
    assert p.is_empty is False
    assert p.path is not None
    assert p.mesh is None
    assert len(list(p.vertices())) == 5, "expected closed path"

    box = p.bbox()
    assert box.extmin == (0, 0)
    assert box.extmax == (1, 1)


def test_poly_face_mesh_to_primitive():
    from ezdxf.layouts import VirtualLayout
    from ezdxf.render.forms import cube

    vl = VirtualLayout()
    poly_face_mesh = cube().render_polyface(vl)
    assert poly_face_mesh.dxftype() == "POLYLINE"

    p = disassemble.make_primitive(poly_face_mesh)
    assert p.is_empty is False
    assert p.path is None
    mesh_builder = p.mesh
    assert mesh_builder is not None

    assert len(mesh_builder.vertices) == 8
    assert len(mesh_builder.faces) == 6
    assert len(list(p.vertices())) == 8

    box = p.bbox()
    assert box.extmin == (-0.5, -0.5, -0.5)
    assert box.extmax == (0.5, 0.5, 0.5)


def test_poly_mesh_to_primitive():
    from ezdxf.layouts import VirtualLayout

    vl = VirtualLayout()
    poly_mesh = vl.add_polymesh(size=(4, 4))
    for x in range(4):
        for y in range(4):
            poly_mesh.set_mesh_vertex((x, y), (x, y, 1.0))

    p = disassemble.make_primitive(poly_mesh)
    assert p.is_empty is False
    assert p.path is None
    mesh_builder = p.mesh
    assert mesh_builder is not None

    assert len(mesh_builder.vertices) == 16
    assert len(mesh_builder.faces) == 9
    assert len(list(p.vertices())) == 16

    box = p.bbox()
    assert box.extmin == (0, 0, 1)
    assert box.extmax == (3, 3, 1)


def test_2d_3d_polyline_to_primitive():
    from ezdxf.layouts import VirtualLayout

    vl = VirtualLayout()

    p1 = Vec3(1, 1)
    p2 = Vec3(2, 2)
    p3 = Vec3(3, 3)
    e2d = vl.add_polyline2d([p1, p2, p3])
    e3d = vl.add_polyline3d([p1, p2, p3])
    for e in [e2d, e3d]:
        p = disassemble.make_primitive(e)
        assert p.is_empty is False
        assert p.path is not None
        assert p.mesh is None
        assert list(p.vertices()) == [p1, p2, p3]

    box = p.bbox()
    assert box.extmin == (1, 1)
    assert box.extmax == (3, 3)


def test_2d_polyline_including_width_to_primitive():
    from ezdxf.layouts import VirtualLayout

    vl = VirtualLayout()

    p1 = (0, 0, 1, 1, 0)
    p2 = (2, 0, 0, 0, 0)
    lwp = vl.add_lwpolyline([p1, p2], dxfattribs={"elevation": 1})
    p2d = vl.add_polyline2d(
        [p1, p2], format="xyseb", dxfattribs={"elevation": (0, 0, 1)}
    )

    for e in [lwp, p2d]:
        p = disassemble.make_primitive(e)
        assert p.is_empty is False
        assert p.path is None
        assert (
            p.mesh is not None
        ), "2D polylines including width should create a mesh"
        vertices = list(p.vertices())
        assert len(vertices) == 4

        box = p.bbox()
        assert box.extmin.isclose((0, -0.5, 1)), "vertices should be in WCS"
        assert box.extmax.isclose((2, 0.5, 1)), "vertices should be in WCS"


def test_text_to_primitive():
    # Testing just the control flow, correct bounding boxes are visually tested.
    # see: ezdxf/examples/entities/text.py
    text = factory.new("TEXT")
    text.dxf.text = "0123456789"
    p = disassemble.make_primitive(text)
    assert p.is_empty is False
    assert p.path is not None
    assert p.mesh is None
    assert len(list(p.vertices())) == 5, "expected closed box"

    box = p.bbox()
    # exact bounding box size depends on available fonts
    assert 18 < box.size.x < 24
    assert box.size.y > 2.5


def test_mtext_to_primitive():
    # Testing just the control flow, correct bounding boxes are visually tested.
    # see: ezdxf/examples/entities/mtext.py
    mtext = factory.new("MTEXT")
    mtext.text = "0123456789"
    p = disassemble.make_primitive(mtext)
    assert p.is_empty is False
    assert p.path is not None
    assert p.mesh is None
    assert len(list(p.vertices())) == 5, "expected closed box"

    box = p.bbox()
    # exact bounding box size depends on available fonts
    assert 18 < box.size.x < 24  # txt.shx
    assert box.size.y > 2.5


def test_mtext_columns_to_primitive():
    from ezdxf.entities.mtext import MTextColumns

    mtext = factory.new("MTEXT")
    mtext._columns = MTextColumns.new_static_columns(3, 10, 1, 15)
    p = disassemble.make_primitive(mtext)
    assert p.is_empty is False
    assert p.path is not None
    assert p.mesh is None
    vertices = list(p.vertices())
    assert len(vertices) == 5, "expected closed box"
    assert vertices[0].isclose((0, 0))
    assert vertices[1].isclose((32, 0))
    assert vertices[2].isclose((32, -15))
    assert vertices[3].isclose((0, -15))

    box = p.bbox()
    assert box.extmin == (0, -15)
    assert box.extmax == (32, 0)


def test_hatch_returns_multiple_primitives():
    hatch = factory.new("HATCH")
    paths = hatch.paths

    # Conversion of boundary paths is tested in 708.
    paths.add_polyline_path([(0, 0), (1, 0), (1, 1)])
    paths.add_polyline_path([(0, 2), (1, 2), (1, 3), (0, 3)])
    res = list(disassemble.to_primitives([hatch]))
    assert len(res) == 1
    p = res[0]
    assert p.path is not None
    assert p.path.has_sub_paths is True
    p0, p1 = p.path.sub_paths()
    v0 = list(p0.approximate())
    v1 = list(p1.approximate())
    assert len(v0) == 4, "expected closed triangle"
    assert len(v1) == 5, "expected closed box"

    box = p.bbox()
    assert box.extmin == (0, 0)
    assert box.extmax == (1, 3)


def test_image_primitive():
    image = factory.new("IMAGE")
    image.dxf.insert = (0, 0)
    image.dxf.u_pixel = Vec3(1, 0)
    image.dxf.v_pixel = Vec3(0, -1)
    image.size = (200, 100)
    image.boundary_path = [(0, 0), (200, 100)]
    p = disassemble.make_primitive(image)
    vertices = list(p.vertices())
    assert len(vertices) == 5, "expected closed box"
    assert vertices[0] == (0.5, -0.5, 0)
    assert vertices[1] == (200.5, -0.5, 0)
    assert vertices[2] == (200.5, 99.5, 0)
    assert vertices[3] == (0.5, 99.5, 0)

    box = p.bbox()
    assert box.extmin == (0.5, -0.5)
    assert box.extmax == (200.5, 99.5)


@pytest.fixture(scope="module")
def circle_primitive():
    circle = factory.new("CIRCLE", dxfattribs={"radius": 3})
    return disassemble.make_primitive(circle)


def test_to_vertices(circle_primitive):
    vertices = list(disassemble.to_vertices([circle_primitive]))
    assert len(vertices) == 40


def test_to_control_vertices(circle_primitive):
    vertices = list(disassemble.to_control_vertices([circle_primitive]))
    # control points from 4 cubic bezier curves:
    assert len(vertices) == 13  # closed: first == last


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Union, Mapping, Optional
import heapq

if TYPE_CHECKING:
    from ezdxf.entities import DXFGraphic

__all__ = ["ascending", "descending"]


def ascending(
    entities: Iterable[DXFGraphic],
    mapping: Optional[Union[dict, Iterable[tuple[str, str]]]] = None,
) -> Iterable[DXFGraphic]:
    """Yields entities in ascending handle order.

    The sort-handle doesn't have to be the entity handle, every entity handle
    in `mapping` will be replaced by the given sort-handle, `mapping` is an
    iterable of 2-tuples (entity_handle, sort_handle) or a
    dict (entity_handle, sort_handle). Entities with equal sort-handles show
    up in source entities order.

    Args:
        entities: iterable of :class:`DXFGraphic` objects
        mapping: iterable of 2-tuples (entity_handle, sort_handle) or a
            handle mapping as dict.

    """
    mapping = dict(mapping) if mapping else {}
    heap = _build(entities, mapping, +1)
    return _sorted(heap)


def descending(
    entities: Iterable[DXFGraphic],
    mapping: Optional[Union[dict, Iterable[tuple[str, str]]]] = None,
) -> Iterable[DXFGraphic]:
    """Yields entities in descending handle order.

    The sort-handle doesn't have to be the entity handle, every entity handle
    in `mapping` will be replaced by the given sort-handle, `mapping` is an
    iterable of 2-tuples (entity_handle, sort_handle) or a
    dict (entity_handle, sort_handle). Entities with equal sort-handles show
    up in reversed source entities order.

    Args:
        entities: iterable of :class:`DXFGraphic` objects
        mapping: iterable of 2-tuples (entity_handle, sort_handle) or a
            handle mapping as dict.

    """
    mapping = dict(mapping) if mapping else {}
    heap = _build(entities, mapping, -1)
    return _sorted(heap)


def _sorted(heap) -> Iterable[DXFGraphic]:
    """Yields heap content in order."""
    while heap:
        yield heapq.heappop(heap)[-1]


def _build(
    entities: Iterable[DXFGraphic], mapping: Mapping, order: int
) -> list[tuple[int, int, DXFGraphic]]:
    """Returns a heap structure.

    Args:
        entities: DXF entities to order
        mapping: handle remapping
        order: +1 for ascending, -1 for descending

    """

    def sort_handle(entity: DXFGraphic) -> int:
        handle = entity.dxf.handle
        sort_handle_ = int(mapping.get(handle, handle), 16)
        # Special handling of sort-handle "0": this behavior is defined by
        # AutoCAD but not documented in the DXF reference.
        # max handle value: ODA DWG Specs: 2.13. Handle References
        # COUNTER is 4 bits, which allows handles up to 16 * 1 byte = 128-bit
        # Example for 128-bit handles: "CADKitSamples\AEC Plan Elev Sample.dxf"
        return sort_handle_ if sort_handle_ else 0xFFFFFFFFFFFFFFFF

    heap: list[tuple[int, int, DXFGraphic]] = []
    for index, entity in enumerate(entities):
        # DXFGraphic is not sortable, using the index as second value avoids
        # a key function and preserves explicit the source order for
        # equal sort-handles.
        heapq.heappush(
            heap,
            (
                sort_handle(entity) * order,
                index * order,
                entity,
            ),
        )
    return heap



# Created: 12.03.2011, 2018 rewritten for pytest
# Copyright (C) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.tools.handle import HandleGenerator


def test_next():
    handles = HandleGenerator("100")
    assert "100" == handles.next()


def test_next_function():
    handles = HandleGenerator("100")
    assert "100" == next(handles)


def test_seed():
    handles = HandleGenerator("200")
    handles.next()
    assert "201" == str(handles)


def test_returns_not_zero():
    handles = HandleGenerator("0")
    assert handles.next() != "0"


def test_returns_not_negative():
    handles = HandleGenerator("-2")
    assert int(handles.next(), 16) > 0


def test_reset():
    handles = HandleGenerator("200")
    handles.reset("300")
    assert "300" == str(handles)


def test_init_and_reset_rejects_invalid_ints():
    with pytest.raises(ValueError):
        HandleGenerator("XXXX")
    handles = HandleGenerator()
    with pytest.raises(ValueError):
        handles.reset("xyz")


def test_copy_handle_generator():
    h0 = HandleGenerator("1")
    h1 = h0.copy()
    assert h0.next() == h1.next()
    h1.next()
    assert str(h0) != str(h1)


def test_copied_handle_generators_are_independent():
    h0 = HandleGenerator("1")
    h1 = h0.copy()
    h1.next()
    assert str(h0) != str(h1)



# Copyright (c) 2010-2023, Manfred Moitzi
# License: MIT License
import pytest
import pickle
from math import radians, sin, cos, pi, isclose
import numpy as np

# Import from 'ezdxf.math._matrix44' to test Python implementation
from ezdxf.math import (
    close_vectors,
    has_matrix_2d_stretching,
    has_matrix_3d_stretching,
)
from ezdxf.math._matrix44 import Matrix44
from ezdxf.acc import USE_C_EXT

m44_classes = [Matrix44]

if USE_C_EXT:
    from ezdxf.acc.matrix44 import Matrix44 as CMatrix44

    m44_classes.append(CMatrix44)


@pytest.fixture(params=m44_classes)
def m44(request):
    return request.param


def diag(values, m44_cls):
    m = m44_cls()
    for i, value in enumerate(values):
        m[i, i] = value
    return m


def equal_matrix(m1, m2, abs_tol=1e-9):
    for row in range(4):
        for col in range(4):
            if not isclose(m1[row, col], m2[row, col], abs_tol=abs_tol):
                return False
    return True


class TestMatrix44:
    @pytest.mark.parametrize("index", [0, 1, 2, 3])
    def test_default_constructor(self, index, m44):
        matrix = m44()
        assert matrix[index, index] == 1.0

    def test_numbers_constructor(self, m44):
        matrix = m44([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15])
        assert matrix.get_row(0) == (0.0, 1.0, 2.0, 3.0)
        assert matrix.get_row(1) == (4.0, 5.0, 6.0, 7.0)
        assert matrix.get_row(2) == (8.0, 9.0, 10.0, 11.0)
        assert matrix.get_row(3) == (12.0, 13.0, 14.0, 15.0)

    def test_row_constructor(self, m44):
        matrix = m44((0, 1, 2, 3), (4, 5, 6, 7), (8, 9, 10, 11), (12, 13, 14, 15))
        assert matrix.get_row(0) == (0.0, 1.0, 2.0, 3.0)
        assert matrix.get_row(1) == (4.0, 5.0, 6.0, 7.0)
        assert matrix.get_row(2) == (8.0, 9.0, 10.0, 11.0)
        assert matrix.get_row(3) == (12.0, 13.0, 14.0, 15.0)

    def test_invalid_row_constructor(self, m44):
        with pytest.raises(ValueError):
            m44((0, 1, 2, 3), (4, 5, 6, 7), (8, 9, 10, 11), (12, 13, 14, 15, 16))
        with pytest.raises(ValueError):
            m44(
                (0, 1, 2, 3),
                (4, 5, 6, 7),
                (8, 9, 10, 11),
                (
                    12,
                    13,
                    14,
                ),
            )

    def test_invalid_number_constructor(self, m44):
        pytest.raises(ValueError, m44, range(17))
        pytest.raises(ValueError, m44, range(15))

    def test_get_item_does_not_support_slicing(self, m44):
        with pytest.raises(TypeError):
            _ = m44()[:]

    def test_get_item_index_error(self, m44):
        with pytest.raises(IndexError):
            _ = m44()[(-1, -1)]
        with pytest.raises(IndexError):
            _ = m44()[(0, 4)]
        with pytest.raises(IndexError):
            _ = m44()[(1, -1)]
        with pytest.raises(IndexError):
            _ = m44()[4, 4]

    def test_set_item_does_not_support_slicing(self, m44):
        with pytest.raises(TypeError):
            m44()[:] = (1, 2)

    def test_set_item_index_error(self, m44):
        with pytest.raises(IndexError):
            m44()[-1, -1] = 0
        with pytest.raises(IndexError):
            m44()[4, 4] = 0

    def test_iter(self, m44):
        values = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
        matrix = m44(values)
        for v1, m1 in zip(values, matrix):
            assert v1 == m1

    def test_copy(self, m44):
        values = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
        m1 = m44(values)
        matrix = m1.copy()
        for v1, m1 in zip(values, matrix):
            assert v1 == m1

    def test_get_row_index_error(self, m44):
        with pytest.raises(IndexError):
            m44().get_row(-1)
        with pytest.raises(IndexError):
            m44().get_row(4)

    def test_set_row(self, m44):
        matrix = m44()
        matrix.set_row(0, (2.0, 3.0, 4.0, 5.0))
        assert matrix.get_row(0) == (2.0, 3.0, 4.0, 5.0)
        matrix.set_row(1, (6.0, 7.0, 8.0, 9.0))
        assert matrix.get_row(1) == (6.0, 7.0, 8.0, 9.0)
        matrix.set_row(2, (10.0, 11.0, 12.0, 13.0))
        assert matrix.get_row(2) == (10.0, 11.0, 12.0, 13.0)
        matrix.set_row(3, (14.0, 15.0, 16.0, 17.0))
        assert matrix.get_row(3) == (14.0, 15.0, 16.0, 17.0)

    def test_set_row_index_error(self, m44):
        with pytest.raises(IndexError):
            m44().set_row(-1, (0,))
        with pytest.raises(IndexError):
            m44().set_row(4, (0,))

    def test_get_col(self, m44):
        matrix = m44()
        assert matrix.get_col(0) == (1.0, 0.0, 0.0, 0.0)
        assert matrix.get_col(1) == (0.0, 1.0, 0.0, 0.0)
        assert matrix.get_col(2) == (0.0, 0.0, 1.0, 0.0)
        assert matrix.get_col(3) == (0.0, 0.0, 0.0, 1.0)

    def test_get_col_index_error(self, m44):
        with pytest.raises(IndexError):
            m44().get_col(-1)
        with pytest.raises(IndexError):
            m44().get_col(4)

    def test_set_col(self, m44):
        matrix = m44()
        matrix.set_col(0, (2.0, 3.0, 4.0, 5.0))
        assert matrix.get_col(0) == (2.0, 3.0, 4.0, 5.0)
        matrix.set_col(1, (6.0, 7.0, 8.0, 9.0))
        assert matrix.get_col(1) == (6.0, 7.0, 8.0, 9.0)
        matrix.set_col(2, (10.0, 11.0, 12.0, 13.0))
        assert matrix.get_col(2) == (10.0, 11.0, 12.0, 13.0)
        matrix.set_col(3, (14.0, 15.0, 16.0, 17.0))
        assert matrix.get_col(3) == (14.0, 15.0, 16.0, 17.0)

    def test_set_col_index_error(self, m44):
        with pytest.raises(IndexError):
            m44().set_col(-1, (0,))
        with pytest.raises(IndexError):
            m44().set_col(4, (0,))

    def test_is_orthogonal(self, m44):
        assert m44().is_orthogonal is True

    def test_is_cartesian(self, m44):
        assert m44().is_cartesian is True

    def test_translate(self, m44):
        t = m44.translate(10, 20, 30)
        x = diag((1.0, 1.0, 1.0, 1.0), m44)
        x[3, 0] = 10.0
        x[3, 1] = 20.0
        x[3, 2] = 30.0
        assert equal_matrix(t, x) is True

    def test_scale(self, m44):
        t = m44.scale(10, 20, 30)
        x = diag((10.0, 20.0, 30.0, 1.0), m44)
        assert equal_matrix(t, x) is True

    def test_x_rotate(self, m44):
        alpha = radians(25)
        t = m44.x_rotate(alpha)
        x = diag((1.0, 1.0, 1.0, 1.0), m44)
        x[1, 1] = cos(alpha)
        x[2, 1] = -sin(alpha)
        x[1, 2] = sin(alpha)
        x[2, 2] = cos(alpha)
        assert equal_matrix(t, x) is True

    def test_y_rotate(self, m44):
        alpha = radians(25)
        t = m44.y_rotate(alpha)
        x = diag((1.0, 1.0, 1.0, 1.0), m44)
        x[0, 0] = cos(alpha)
        x[2, 0] = sin(alpha)
        x[0, 2] = -sin(alpha)
        x[2, 2] = cos(alpha)
        assert equal_matrix(t, x) is True

    def test_z_rotate(self, m44):
        alpha = radians(25)
        t = m44.z_rotate(alpha)
        x = diag((1.0, 1.0, 1.0, 1.0), m44)
        x[0, 0] = cos(alpha)
        x[1, 0] = -sin(alpha)
        x[0, 1] = sin(alpha)
        x[1, 1] = cos(alpha)
        assert equal_matrix(t, x) is True

    def test_chain(self, m44):
        s = m44.scale(10, 20, 30)
        t = m44.translate(10, 20, 30)

        c = m44.chain(s, t)
        x = diag((10.0, 20.0, 30.0, 1.0), m44)
        x[3, 0] = 10.0
        x[3, 1] = 20.0
        x[3, 2] = 30.0
        assert equal_matrix(c, x) is True

    def test_chain2(self, m44):
        s = m44.scale(10, 20, 30)
        t = m44.translate(10, 20, 30)
        r = m44.axis_rotate(angle=pi / 2, axis=(0.0, 0.0, 1.0))
        points = ((23.0, 97.0, 0.5), (2.0, 7.0, 13.0))

        p1 = s.transform_vertices(points)
        p1 = t.transform_vertices(p1)
        p1 = r.transform_vertices(p1)

        c = m44.chain(s, t, r)
        p2 = c.transform_vertices(points)
        assert close_vectors(p1, p2) is True

    def test_transform(self, m44):
        t = m44.scale(2.0, 0.5, 1.0)
        r = t.transform((10.0, 20.0, 30.0))
        assert r == (20.0, 10.0, 30.0)

    def test_multiply(self, m44):
        m1 = m44(range(16))
        m2 = m44(range(16))
        res = m1 * m2
        expected = m44(
            (56.0, 62.0, 68.0, 74.0),
            (152.0, 174.0, 196.0, 218.0),
            (248.0, 286.0, 324.0, 362.0),
            (344.0, 398.0, 452.0, 506.0),
        )
        assert equal_matrix(res, expected)
        # __matmul__()
        res = m1 @ m2
        assert equal_matrix(res, expected)

    def test_transpose(self, m44):
        matrix = m44((0, 1, 2, 3), (4, 5, 6, 7), (8, 9, 10, 11), (12, 13, 14, 15))
        matrix.transpose()
        assert matrix.get_row(0) == (0.0, 4.0, 8.0, 12.0)
        assert matrix.get_row(1) == (1.0, 5.0, 9.0, 13.0)
        assert matrix.get_row(2) == (2.0, 6.0, 10.0, 14.0)
        assert matrix.get_row(3) == (3.0, 7.0, 11.0, 15.0)

    def test_inverse_error(self, m44):
        m = m44([1] * 16)
        pytest.raises(ZeroDivisionError, m.inverse)

    def test_determinant(self, m44):
        matrix = m44((3, 1, 3, 3), (5, 5, 6, 7), (8, 9, 0, 11), (12, 14, 14, 15))
        assert isclose(matrix.determinant(), 289.)

    def test_axis_rotate_for_axis_normalization(self, m44):
        m1 = m44.axis_rotate((0, 0, 1), 1.23)
        m2 = m44.axis_rotate((0, 0, 0.5), 1.23)
        for a, b in zip(m1, m2):
            assert isclose(a, b)

    def test_assign_after_initialized(self, m44):
        matrix = m44()
        matrix[0, 0] = 12
        matrix2 = m44()
        assert matrix2[0, 0] == 1

        values = list(range(16))
        matrix = m44(values)
        matrix[0, 0] = 12
        assert values[0] == 0
        assert matrix[0, 0] == 12

    def test_picklable(self, m44):
        matrix = m44((0.1, 1, 2, 3), (4, 5, 6, 7), (8, 9, 10, 11), (12, 13, 14, 15))
        pickled_matrix = pickle.loads(pickle.dumps(matrix))
        assert equal_matrix(matrix, pickled_matrix)
        assert type(matrix) is type(pickled_matrix)
        matrix[0, 0] = 12
        assert not equal_matrix(matrix, pickled_matrix)

    def test_shear_xy(self, m44):
        angle = pi / 4
        matrix = m44.shear_xy(angle_x=angle, angle_y=-angle)
        assert matrix[0, 1] == pytest.approx(-1.0)
        assert matrix[1, 0] == pytest.approx(1.0)

    def test_from_2d_transformation(self, m44):
        components = (2, 0, 0, 2, 10, 20)
        matrix = m44.from_2d_transformation(components)
        assert matrix.transform((1, 1)).isclose((12, 22))

    @pytest.mark.parametrize(
        "components",
        [
            [],
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4, 5, 6],
        ],
    )
    def test_from_2d_transformation_checks_component_count(self, m44, components):
        with pytest.raises(ValueError):
            m44.from_2d_transformation(components)


def test_has_matrix_2d_stretching():
    """Note: Uniform scaling is not stretching in this context."""
    assert has_matrix_2d_stretching(Matrix44.scale(1, 1, 1)) is False
    assert has_matrix_2d_stretching(Matrix44.scale(2, 2, 2)) is False
    assert has_matrix_2d_stretching(Matrix44.scale(1, 1, 2)) is False, "ignore z-axis"
    assert has_matrix_2d_stretching(Matrix44.scale(2, 1, 1)) is True
    assert has_matrix_2d_stretching(Matrix44.scale(1, 2, 1)) is True


def test_has_matrix_3d_stretching():
    """Note: Uniform scaling is not stretching in this context."""
    assert has_matrix_3d_stretching(Matrix44.scale(1, 1, 1)) is False
    assert has_matrix_3d_stretching(Matrix44.scale(2, 2, 2)) is False
    assert has_matrix_3d_stretching(Matrix44.scale(2, 1, 1)) is True
    assert has_matrix_3d_stretching(Matrix44.scale(1, 2, 1)) is True
    assert has_matrix_3d_stretching(Matrix44.scale(1, 1, 2)) is True


class TestFast2dTransform:
    def test_fast_translate(self, m44):
        m = m44.translate(10, 20, 0)
        points = list(m.fast_2d_transform([(0, 0), (1, 1, 1)]))
        assert points[0].isclose((10, 20))
        assert points[1].isclose((11, 21))

    def test_fast_z_rotate(self, m44):
        m = m44.z_rotate(radians(90))
        points = list(m.fast_2d_transform([(10, 0), (0, 10, 1)]))
        assert points[0].isclose((0, 10))
        assert points[1].isclose((-10, 0))

    def test_fast_scale(self, m44):
        m = m44.scale(2, 3, 4)
        points = list(m.fast_2d_transform([(10, 10), (-20, -20, 1)]))
        assert points[0].isclose((20, 30))
        assert points[1].isclose((-40, -60))

    def test_fast_scale_rotate_translate(self, m44):
        m = m44.scale(2, 3, 4) @ m44.z_rotate(radians(90)) @ m44.translate(10, 20, 0)
        points = [(10, 10), (-20, -20, 1)]
        res_3d = list(m.transform_vertices(points))
        res_2d = list(m.fast_2d_transform(points))

        assert res_2d[0].isclose(res_3d[0])
        assert res_2d[1].isclose(res_3d[1])


class TestTransformArrayInplace:
    def test_ndim_2(self, m44):
        points = (23.0, 97.0), (2.0, 7.0)
        s = m44.scale(10, 20, 1)
        t = m44.translate(10, 20, 0)
        r = m44.z_rotate(angle=pi / 2)
        m = m44.chain(s, r, t)
        array = np.array(points, dtype=np.float64)

        control = list(m.fast_2d_transform(points))
        m.transform_array_inplace(array, ndim=2)
        assert close_vectors(control, array) is True

    def test_ndim_3(self, m44):
        points = (23.0, 97.0, 0.5), (2.0, 7.0, 13.0)
        s = m44.scale(10, 20, 30)
        t = m44.translate(10, 20, 30)
        r = m44.z_rotate(angle=pi / 2)
        m = m44.chain(s, r, t)
        array = np.array(points, dtype=np.float64)

        control = list(m.transform_vertices(points))
        m.transform_array_inplace(array, ndim=3)
        assert close_vectors(control, array) is True



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import NamedTuple, Sequence


class Command(NamedTuple):
    name: str
    args: Sequence[bytes]


ESCAPE = 27
SEMICOLON = ord(";")
PERCENT = ord("%")
MINUS = ord("-")
QUOTE_CHAR = ord('"')
CHAR_A = ord("A")
CHAR_B = ord("B")
CHAR_Z = ord("Z")
DEFAULT_TEXT_TERMINATOR = 3

# Enter HPGL/2 mode commands
# b"%-0B",  # ??? not documented (assumption)
# b"%-1B",  # ??? not documented (really exist)
# b"%-2B",  # ??? not documented (assumption)
# b"%-3B",  # ??? not documented (assumption)
# b"%0B",  # documented in the HPGL2 reference by HP
# b"%1B",  # documented
# b"%2B",  # documented
# b"%3B",  # documented


def get_enter_hpgl2_mode_command_length(s: bytes, i: int) -> int:
    try:
        if s[i] != ESCAPE:
            return 0
        if s[i + 1] != PERCENT:
            return 0
        length = 4
        if s[i + 2] == MINUS:
            i += 1
            length = 5
        # 0, 1, 2 or 3 + "B"
        if 47 < s[i + 2] < 52 and s[i + 3] == CHAR_B:
            return length
    except IndexError:
        pass
    return 0


KNOWN_START_SEQUENCES = [b"BPIN", b"BP;IN", b"INPS", b"IN;PS", b"INDF", b"IN;DF"]


def has_known_start_sequence(b: bytes) -> bool:
    for start_sequence in KNOWN_START_SEQUENCES:
        if b.startswith(start_sequence):
            return True
    return False


def find_hpgl2_entry_point(s: bytes, start: int) -> int:
    while True:
        try:
            index = s.index(b"%", start)
        except ValueError:
            return len(s)
        length = get_enter_hpgl2_mode_command_length(s, index)
        if length:
            return index + length
        start += 2


def hpgl2_commands(s: bytes) -> list[Command]:
    """Low level plot file parser, extracts the HPGL/2 from the byte stream `b`.

    .. Important::

        This parser expects the "Enter HPGL/2 mode" escape sequence to recognize
        HPGL/2 commands. The sequence looks like this: ``[ESC]%1B``, multiple variants
        of this sequence are supported.

    """
    text_terminator = DEFAULT_TEXT_TERMINATOR

    def find_terminator(i: int) -> int:
        while i < length:
            c = s[i]
            if c == QUOTE_CHAR:
                i = find_mark(i + 1, QUOTE_CHAR)
            elif (CHAR_A <= c <= CHAR_Z) or c == SEMICOLON or c == ESCAPE:
                break
            i += 1
        return i

    def find_mark(i: int, mark: int) -> int:
        while i < length and s[i] != mark:
            i += 1
        return i + 1

    def append_command(b: bytes) -> None:
        if b[:2] == b"DT":
            nonlocal text_terminator
            if len(b) > 2:
                text_terminator = b[2]
            else:
                text_terminator = DEFAULT_TEXT_TERMINATOR
        else:
            commands.append(make_command(b))

    commands: list[Command] = []
    length = len(s)
    if has_known_start_sequence(s):
        index = 0
    else:
        index = find_hpgl2_entry_point(s, 0)
    while index < length:
        char = s[index]
        start = index

        if char == ESCAPE:
            # HPGL/2 does not use escape sequences, whatever this sequence is,
            # HPGL/2 mode was left. Find next entry point into HPGL/2 mode:
            index = find_hpgl2_entry_point(s, index)
            continue

        if char <= 32:  # skip all white space and control chars between commands
            index += 1
            continue

        index_plus_2 = index + 2
        if index_plus_2 >= length:
            append_command(s[index:])
            break

        command = s[start:index_plus_2]

        if command == b"PE":
            index = find_mark(index_plus_2, SEMICOLON)
            index -= 1  # exclude terminator ";" from command args
        elif command == b"LB":
            index = find_mark(index_plus_2, text_terminator)
            # include special terminator in command args,
            # otherwise the parser is confused
        else:
            index = find_terminator(index_plus_2)

        append_command(s[start:index])
        if index < length and s[index] == SEMICOLON:
            index += 1
    return commands


def make_command(cmd: bytes) -> Command:
    if not cmd:
        return Command("NOOP", tuple())
    name = cmd[:2].decode()
    if name == "PE":
        args = (bytes([c for c in cmd[2:] if c > 32]),)
    else:
        args = tuple(s for s in cmd[2:].split(b","))  # type: ignore
    return Command(name, args)


def fractional_bits(decimal_places: int) -> int:
    return round(decimal_places * 3.33)


def pe_encode(value: float, frac_bits: int = 0, base: int = 64) -> bytes:
    if frac_bits:
        value *= 1 << frac_bits
        x = round(value)
    else:
        x = round(value)
    if x >= 0:
        x *= 2
    else:
        x = abs(x) * 2 + 1

    chars = bytearray()
    while x >= base:
        x, r = divmod(x, base)
        chars.append(63 + r)
    if base == 64:
        chars.append(191 + x)
    else:
        chars.append(95 + x)
    return bytes(chars)


def pe_decode(
    s: bytes, frac_bits: int = 0, base=64, start: int = 0
) -> tuple[list[float], int]:
    def _decode():
        factors.reverse()
        x = 0
        for f in factors:
            x = x * base + f
        factors.clear()
        if x & 1:
            x = -(x - 1)
        x = x >> 1
        return x

    n = 1 << frac_bits
    if base == 64:
        terminator = 191
    else:
        terminator = 95
    values: list[float] = []
    factors = []
    for index in range(start, len(s)):
        value = s[index]
        if value < 63:
            return values, index
        if value >= terminator:
            factors.append(value - terminator)
            x = _decode()
            if frac_bits:
                values.append(x / n)
            else:
                values.append(float(x))
        else:
            factors.append(value - 63)
    return values, len(s)



# Copyright (C) 2018-2019, Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.tools.complex_ltype import (
    lin_tokenizer,
    lin_parser,
    lin_compiler,
    ComplexLineTypePart,
)


def test_line_type_tokenizer_just_numbers():
    ltype = "A,.25,-.125,.25,-.125,0,-.125"
    result = list(lin_tokenizer(ltype))
    assert result == ["A", ".25", "-.125", ".25", "-.125", "0", "-.125"]


def test_line_type_tokenizer_strings():
    ltype = 'A,.5,-.2,["GAS",STANDARD,S=.1,U=0.0,X=-0.1,Y=-.05],-.25'
    result = list(lin_tokenizer(ltype))
    assert result == [
        "A",
        ".5",
        "-.2",
        '["GAS"',
        "STANDARD",
        "S=.1",
        "U=0.0",
        "X=-0.1",
        "Y=-.05]",
        "-.25",
    ]


def test_line_type_tokenizer_string_with_comma():
    ltype = 'A, "TEXT, TEXT", 0'
    result = list(lin_tokenizer(ltype))
    assert result == ["A", '"TEXT, TEXT"', "0"]


def test_line_type_tokenizer_shapes():
    # A,.25,-.1,[BOX,ltypeshp.shx,x=-.1,s=.1],-.1,1 replacing BOX by shape index 132
    ltype = "A,.25,-.1,[132,ltypeshp.shx,x=-.1,s=.1],-.1,1"
    result = list(lin_tokenizer(ltype))
    assert result == [
        "A",
        ".25",
        "-.1",
        "[132",
        "ltypeshp.shx",
        "x=-.1",
        "s=.1]",
        "-.1",
        "1",
    ]


def test_line_type_parser_just_numbers():
    ltype = "A,.25,-.125,.25,-.125,0,-.125"
    result = lin_parser(ltype)
    assert result == ["A", 0.25, -0.125, 0.25, -0.125, 0, -0.125]


def test_line_type_parser_strings():
    ltype = 'A,.5,-.2,["GAS",STANDARD,S=.1,U=0.0,X=-0.1,Y=-.05],-.25'
    result = lin_parser(ltype)
    assert result == [
        "A",
        0.5,
        -0.2,
        ["TEXT", "GAS", "STANDARD", "s", 0.1, "u", 0.0, "x", -0.1, "y", -0.05],
        -0.25,
    ]


def test_line_type_parser_shape():
    # A,.25,-.1,[BOX,ltypeshp.shx,x=-.1,s=.1],-.1,1 replacing BOX by shape index 132
    ltype = "A,.25,-.1,[132,ltypeshp.shx,x=-.1,s=.1],-.1,1"
    result = lin_parser(ltype)
    assert result == [
        "A",
        0.25,
        -0.1,
        ["SHAPE", 132, "ltypeshp.shx", "x", -0.1, "s", 0.1],
        -0.1,
        1,
    ]


def test_lin_compiler_floats():
    ltype = "A,.25,-.125,.25,-.125,0,-.125"
    result = lin_compiler(ltype)
    assert result == [
        (49, 0.25),
        (49, -0.125),
        (49, 0.25),
        (49, -0.125),
        (49, 0),
        (49, -0.125),
    ]


def test_lin_compiler_strings():
    ltype = 'A,.5,-.2,["GAS",STANDARD,S=.1,U=0.0,X=-0.1,Y=-.05],-.25'
    result = lin_compiler(ltype)
    assert result[0] == (49, 0.5)
    assert result[1] == (49, -0.2)
    text = result[2]
    assert text.type == "TEXT"
    assert text.value == "GAS"
    assert text.font == "STANDARD"
    assert text.tags[0] == (46, 0.1)
    assert text.tags[1] == (50, 0)
    assert text.tags[2] == (44, -0.1)
    assert text.tags[3] == (45, -0.05)
    assert result[3] == (49, -0.25)


def test_lin_compiler_shape():
    # A,.25,-.1,[BOX,ltypeshp.shx,x=-.1,s=.1],-.1,1 replacing BOX by shape index 132
    ltype = "A,.25,-.1,[132,ltypeshp.shx,x=-.1,s=.1],-.1,1"
    result = lin_compiler(ltype)
    assert result[0] == (49, 0.25)
    assert result[1] == (49, -0.1)
    text = result[2]
    assert text.type == "SHAPE"
    assert text.value == 132
    assert text.font == "ltypeshp.shx"
    assert text.tags[0] == (46, 0.1)
    assert text.tags[1] == (50, 0.0)
    assert text.tags[2] == (44, -0.1)
    assert text.tags[3] == (45, 0.0)
    assert result[3] == (49, -0.1)
    assert result[4] == (49, 1)


def test_tags_from_complex_text():
    ltype = 'A,.5,-.2,["GAS",STANDARD,S=.1,U=0.0,X=-0.1,Y=-.05],-.25'
    result = lin_compiler(ltype)
    cx_part = result[2]
    assert isinstance(cx_part, ComplexLineTypePart)
    assert cx_part.type == "TEXT"
    tags = cx_part.complex_ltype_tags(None)
    assert tags[0] == (74, 2)
    assert tags[1] == (75, 0)
    assert tags[2] == (340, "0")  # default handle without a DXF document
    assert tags[3] == (46, 0.1)  # s
    assert tags[4] == (50, 0)  # r
    assert tags[5] == (44, -0.1)  # x
    assert tags[6] == (45, -0.05)  # y
    assert tags[7] == (9, "GAS")


def test_tags_from_complex_shape():
    ltype = "A,.25,-.1,[132,ltypeshp.shx,x=-.1,s=.1],-.1,1"
    result = lin_compiler(ltype)
    cx_part = result[2]
    assert isinstance(cx_part, ComplexLineTypePart)
    assert cx_part.type == "SHAPE"
    tags = cx_part.complex_ltype_tags(None)
    assert tags[0] == (74, 4)
    assert tags[1] == (75, 132)  # shape index
    assert tags[2] == (340, "0")  # default handle with a DXF document
    assert tags[3] == (46, 0.1)  # s
    assert tags[4] == (50, 0.0)  # r
    assert tags[5] == (44, -0.1)  # x
    assert tags[6] == (45, 0.0)  # y


def test_new_table_entry_has_text_style_handle():
    import ezdxf

    doc = ezdxf.new("R2013")
    ltype = doc.linetypes.new(
        "GASLEITUNG2",
        dxfattribs={
            "description": "Gasleitung2 ----GAS----GAS----GAS----GAS----GAS----GAS--",
            "length": 1,
            "pattern": 'A,.5,-.2,["GAS",STANDARD,S=.1,U=0.0,X=-0.1,Y=-.05],-.25',
        },
    )
    style_handle = ltype.pattern_tags.get_style_handle()
    assert style_handle == doc.styles.get("Standard").dxf.handle


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2011-2022, Manfred Moitzi
# License: MIT License
from .base import BaseLayout, VirtualLayout
from .layout import Layout, Modelspace, Paperspace
from .blocklayout import BlockLayout
from .layouts import Layouts



# Copyright (c) 2021-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import BinaryIO, Optional, Sequence
from ezdxf.lldxf.validator import is_dxf_file, DXFStructureError
from pathlib import Path


class TagWriter:
    def __init__(self, fp: BinaryIO):
        self.fp = fp

    def write(self, raw_code_str: bytes, raw_value_str: bytes):
        self.fp.write(raw_code_str)
        self.fp.write(raw_value_str)


class ThumbnailRemover(TagWriter):
    def __init__(self, fp: BinaryIO):
        super().__init__(fp)
        self._start_section = False
        self._skip_tags = False
        self._section_code: Optional[bytes] = None
        self._section_value: Optional[bytes] = None
        self.removed_thumbnail_image = False

    def write(self, raw_code_str: bytes, raw_value_str: bytes):
        code = raw_code_str.strip()
        value = raw_value_str.strip()
        if self._start_section:
            self._start_section = False
            if code == b"2" and value == b"THUMBNAILIMAGE":
                self._skip_tags = True
                self.removed_thumbnail_image = True
            else:
                # write buffered section tag:
                super().write(self._section_code, self._section_value)  # type: ignore

        if code == b"0":
            if value == b"SECTION":
                self._start_section = True
                self._skip_tags = False
                # buffer section tag:
                self._section_code = raw_code_str
                self._section_value = raw_value_str
                return
            elif value == b"ENDSEC":
                skip = self._skip_tags
                self._skip_tags = False
                if skip:  # don't write ENDSEC
                    return

        if not self._skip_tags:
            super().write(raw_code_str, raw_value_str)


def strip_tags(
    infile: BinaryIO,
    tagwriter: TagWriter,
    codes: Sequence[int],
    verbose=False,
) -> int:
    search_codes = set(codes)
    line_number: int = 1
    removed_tags: int = 0
    while True:
        try:
            raw_code_str = infile.readline()
        except EOFError:
            raw_code_str = b""
        if raw_code_str == b"":  # regular end of file
            return removed_tags
        try:
            code = int(raw_code_str)
        except ValueError:
            code_str = raw_code_str.strip().decode(encoding="utf8", errors="ignore")
            raise DXFStructureError(
                f'CANCELED: "{infile.name}" - found invalid '
                f'group code "{code_str}" at line {line_number}'
            )

        try:
            raw_value_str = infile.readline()
        except EOFError:
            raw_value_str = b""

        if raw_value_str == b"":
            raise DXFStructureError(
                f'CANCELED: "{infile.name}" - premature end of file'
            )
        line_number += 2
        if code not in search_codes:
            tagwriter.write(raw_code_str, raw_value_str)
        else:
            if verbose:
                value = raw_value_str.strip()
                _value = value.decode(encoding="utf8", errors="ignore")
                print(f'removing tag: ({code}, "{_value}")')
            removed_tags += 1


def safe_rename(source: Path, target: Path, backup=True, verbose=False) -> bool:
    backup_file = target.with_suffix(".bak")
    backup_file.unlink(missing_ok=True)
    _target = Path(target)
    if _target.exists():
        if verbose:
            print(f'renaming "{_target.name}" to "{backup_file.name}"')
        try:
            _target.rename(backup_file)
        except IOError as e:
            print(f"IOError: {str(e)}")
            return False

    if verbose:
        print(f'renaming "{source.name}" to "{target.name}"')
    try:
        source.rename(target)
    except IOError as e:
        print(f"IOError: {str(e)}")
        return False

    if not backup:
        if verbose:
            print(f'deleting backup file "{backup_file.name}"')
        backup_file.unlink(missing_ok=True)
    return True


DEFAULT_CODES = (999,)


def strip(
    filename: str,
    backup=False,
    thumbnail=False,
    verbose=False,
    codes: Sequence[int] = DEFAULT_CODES,
):
    def remove_tmp_file():
        if tmp_file.exists():
            if verbose:
                print(f'deleting temp file: "{tmp_file.name}"')
            tmp_file.unlink(missing_ok=True)

    if verbose:
        print(f'\nProcessing file: "{filename}"')
    try:
        if not is_dxf_file(filename):
            print(
                f'CANCELED: "{filename}" is not a DXF file, binary DXF files '
                f"are not supported"
            )
            return
    except IOError as e:
        print(f"IOError: {str(e)}")
        return
    source_file = Path(filename)
    tmp_file = source_file.with_suffix(".ezdxf.tmp")
    error = False
    tagwriter: TagWriter
    if verbose:
        print(f'make a temporary copy: "{tmp_file.name}"')
    with open(tmp_file, "wb") as fp, open(source_file, "rb") as infile:
        if thumbnail:
            tagwriter = ThumbnailRemover(fp)
        else:
            tagwriter = TagWriter(fp)
        try:
            removed_tags = strip_tags(infile, tagwriter, codes=codes, verbose=verbose)
        except IOError as e:
            print(f"IOError: {str(e)}")
            error = True
        except DXFStructureError as e:
            print(str(e))
            error = True

    if not error:
        rename = False
        if thumbnail and tagwriter.removed_thumbnail_image:  # type: ignore
            print(f'"{source_file.name}" - removed THUMBNAILIMAGE section')
            rename = True

        if removed_tags > 0:
            tags = "tag" if removed_tags == 1 else "tags"
            print(f'"{source_file.name}" - {removed_tags} {tags} removed')
            rename = True

        if rename:
            safe_rename(tmp_file, source_file, backup, verbose)

    remove_tmp_file()



# Copyright (c) 2020-2023, Manfred Moitzi
# License: MIT License
from __future__ import annotations
import logging
from typing import (
    TYPE_CHECKING,
    Iterable,
    Callable,
    Optional,
    cast,
)
from ezdxf.lldxf import const
from ezdxf.entities import factory
from ezdxf.entities.boundary_paths import (
    PolylinePath,
    EdgePath,
    LineEdge,
    ArcEdge,
    EllipseEdge,
    SplineEdge,
)
from ezdxf.math import OCS, Vec3, ABS_TOL
from ezdxf.math.transformtools import (
    NonUniformScalingError,
    InsertTransformationError,
)
from ezdxf.query import EntityQuery
from ezdxf.entities.copy import default_copy, CopyNotSupported

if TYPE_CHECKING:
    from ezdxf.entities import (
        DXFGraphic,
        Insert,
        Attrib,
        Text,
        Dimension,
    )
    from ezdxf.entities.polygon import DXFPolygon
    from ezdxf.layouts import BaseLayout

logger = logging.getLogger("ezdxf")

__all__ = [
    "virtual_block_reference_entities",
    "virtual_boundary_path_entities",
    "explode_block_reference",
    "explode_entity",
    "attrib_to_text",
]


def default_logging_callback(entity, reason):
    logger.debug(
        f'(Virtual Block Reference Entities) Ignoring {str(entity)}: "{reason}"'
    )


def explode_block_reference(
    block_ref: Insert,
    target_layout: BaseLayout,
    *,
    redraw_order=False,
    copy_strategy=default_copy,
) -> EntityQuery:
    """Explode a block reference into DXF primitives.

    Transforms the block entities into the required WCS location by applying the
    block reference attributes `insert`, `extrusion`, `rotation` and the scaling
    values `xscale`, `yscale` and `zscale`.

    Returns an EntityQuery() container with all exploded DXF entities.

    Attached ATTRIB entities are converted to TEXT entities, this is the
    behavior of the BURST command of the AutoCAD Express Tools.

    This method does not apply the clipping path created by the XCLIP command. 
    The method returns all entities and ignores the clipping path polygon and no 
    entity is clipped.

    Args:
        block_ref: Block reference entity (INSERT)
        target_layout: explicit target layout for exploded DXF entities
        redraw_order: create entities in ascending redraw order if ``True``
        copy_strategy: customizable copy strategy

    .. warning::

        **Non uniform scaling** may lead to incorrect results for text entities
        (TEXT, MTEXT, ATTRIB) and maybe some other entities.

    (internal API)

    """
    if target_layout is None:
        raise const.DXFStructureError("Target layout is None.")

    if block_ref.doc is None:
        raise const.DXFStructureError(
            "Block reference has to be assigned to a DXF document."
        )

    def _explode_single_block_ref(block_ref):
        for entity in virtual_block_reference_entities(
            block_ref, redraw_order=redraw_order, copy_strategy=copy_strategy
        ):
            dxftype = entity.dxftype()
            target_layout.add_entity(entity)
            if dxftype == "DIMENSION":
                # Render a graphical representation for each exploded DIMENSION
                # entity as anonymous block.
                cast("Dimension", entity).render()
            entities.append(entity)

        # Convert attached ATTRIB entities to TEXT entities:
        # This is the behavior of the BURST command of the AutoCAD Express Tools
        for attrib in block_ref.attribs:
            # Attached ATTRIB entities are already located in the WCS
            text = attrib_to_text(attrib)
            target_layout.add_entity(text)
            entities.append(text)

    entitydb = block_ref.doc.entitydb
    assert (
        entitydb is not None
    ), "Exploding a block reference requires an entity database."

    entities: list[DXFGraphic] = []
    if block_ref.mcount > 1:
        for virtual_insert in block_ref.multi_insert():
            _explode_single_block_ref(virtual_insert)
    else:
        _explode_single_block_ref(block_ref)

    source_layout = block_ref.get_layout()
    if source_layout is not None:
        # Remove and destroy exploded INSERT if assigned to a layout
        source_layout.delete_entity(block_ref)
    else:
        entitydb.delete_entity(block_ref)
    return EntityQuery(entities)


IGNORE_FROM_ATTRIB = {
    "version",
    "prompt",
    "tag",
    "flags",
    "field_length",
    "lock_position",
    "attribute_type",
}


def attrib_to_text(attrib: Attrib) -> Text:
    dxfattribs = attrib.dxfattribs(drop=IGNORE_FROM_ATTRIB)
    # ATTRIB has same owner as INSERT but does not reside in any EntitySpace()
    # and must not deleted from any layout.
    # New TEXT entity has same handle as the replaced ATTRIB entity and replaces
    # the ATTRIB entity in the database.
    text = factory.new("TEXT", dxfattribs=dxfattribs)
    if attrib.doc:
        factory.bind(text, attrib.doc)
    return cast("Text", text)


def virtual_block_reference_entities(
    block_ref: Insert,
    *,
    skipped_entity_callback: Optional[Callable[[DXFGraphic, str], None]] = None,
    redraw_order=False,
    copy_strategy=default_copy,
) -> Iterable[DXFGraphic]:
    """Yields 'virtual' parts of block reference `block_ref`.

    This method is meant to examine the block reference entities without the need to
    explode the block reference. The `skipped_entity_callback()` will be called for all
    entities which are not processed, signature:
    :code:`skipped_entity_callback(entity: DXFGraphic, reason: str)`,
    `entity` is the original (untransformed) DXF entity of the block definition,
    the `reason` string is an explanation why the entity was skipped.

    These entities are located at the 'exploded' positions, but are not stored in
    the entity database, have no handle and are not assigned to any layout.

    This method does not apply the clipping path created by the XCLIP command. 
    The method returns all entities and ignores the clipping path polygon and no 
    entity is clipped.

    Args:
        block_ref: Block reference entity (INSERT)
        skipped_entity_callback: called whenever the transformation of an entity
            is not supported and so was skipped.
        redraw_order: yield entities in ascending redraw order if ``True``
        copy_strategy: customizable copy strategy

    .. warning::

        **Non uniform scaling** may lead to incorrect results for text entities
        (TEXT, MTEXT, ATTRIB) and maybe some other entities.

    (internal API)

    """
    assert block_ref.dxftype() == "INSERT"
    from ezdxf.entities import Ellipse

    skipped_entity_callback = skipped_entity_callback or default_logging_callback

    def disassemble(layout: BaseLayout) -> Iterable[DXFGraphic]:
        for entity in layout.entities_in_redraw_order() if redraw_order else layout:
            # Do not explode ATTDEF entities. Already available in Insert.attribs
            if entity.dxftype() == "ATTDEF":
                continue
            try:
                copy = entity.copy(copy_strategy=copy_strategy)
            except CopyNotSupported:
                if hasattr(entity, "virtual_entities"):
                    yield from entity.virtual_entities()
                else:
                    skipped_entity_callback(entity, "non copyable")
            else:
                if hasattr(copy, "remove_association"):
                    copy.remove_association()
                yield copy

    def transform(entities):
        for entity in entities:
            try:
                entity.transform(m)
            except NotImplementedError:
                skipped_entity_callback(entity, "non transformable")
            except NonUniformScalingError:
                dxftype = entity.dxftype()
                if dxftype in {"ARC", "CIRCLE"}:
                    if abs(entity.dxf.radius) > ABS_TOL:
                        yield Ellipse.from_arc(entity).transform(m)
                    else:
                        skipped_entity_callback(
                            entity, f"Invalid radius in entity {str(entity)}."
                        )
                elif dxftype in {"LWPOLYLINE", "POLYLINE"}:  # has arcs
                    yield from transform(entity.virtual_entities())
                else:
                    skipped_entity_callback(entity, "unsupported non-uniform scaling")
            except InsertTransformationError:
                # INSERT entity can not be represented in the target coordinate
                # system defined by transformation matrix `m`.
                # Yield transformed sub-entities of the INSERT entity:
                yield from transform(
                    virtual_block_reference_entities(
                        entity, skipped_entity_callback=skipped_entity_callback
                    )
                )
            else:
                yield entity

    m = block_ref.matrix44()
    block_layout = block_ref.block()
    if block_layout is None:
        raise const.DXFStructureError(
            f'Required block definition for "{block_ref.dxf.name}" does not exist.'
        )

    yield from transform(disassemble(block_layout))


EXCLUDE_FROM_EXPLODE = {"POINT"}


def explode_entity(
    entity: DXFGraphic, target_layout: Optional[BaseLayout] = None
) -> EntityQuery:
    """Explode parts of an entity as primitives into target layout, if target
    layout is ``None``, the target layout is the layout of the source entity.

    Returns an :class:`~ezdxf.query.EntityQuery` container with all DXF parts.

    Args:
        entity: DXF entity to explode, has to have a :meth:`virtual_entities()`
            method
        target_layout: target layout for DXF parts, ``None`` for same layout as
            source entity

    (internal API)

    """
    dxftype = entity.dxftype()
    virtual_entities = getattr(entity, "virtual_entities")
    if virtual_entities is None or dxftype in EXCLUDE_FROM_EXPLODE:
        raise const.DXFTypeError(f"Can not explode entity {dxftype}.")

    if entity.doc is None:
        raise const.DXFStructureError(
            f"{dxftype} has to be assigned to a DXF document."
        )

    entitydb = entity.doc.entitydb
    if entitydb is None:
        raise const.DXFStructureError(
            f"Exploding {dxftype} requires an entity database."
        )

    if target_layout is None:
        target_layout = entity.get_layout()
        if target_layout is None:
            raise const.DXFStructureError(
                f"{dxftype} without layout assignment, specify target layout."
            )

    entities = []
    for e in virtual_entities():
        target_layout.add_entity(e)
        entities.append(e)

    source_layout = entity.get_layout()
    if source_layout is not None:
        source_layout.delete_entity(entity)
    else:
        entitydb.delete_entity(entity)
    return EntityQuery(entities)


def virtual_boundary_path_entities(
    polygon: DXFPolygon,
) -> list[list[DXFGraphic]]:
    from ezdxf.entities import LWPolyline

    def polyline():
        p = LWPolyline.new(dxfattribs=dict(graphic_attribs))
        p.append_formatted_vertices(path.vertices, format="xyb")
        p.dxf.extrusion = ocs.uz
        p.dxf.elevation = elevation
        p.closed = path.is_closed
        return p

    graphic_attribs = polygon.graphic_properties()
    elevation = float(polygon.dxf.elevation.z)
    ocs = polygon.ocs()
    entities = []
    for path in polygon.paths:
        if isinstance(path, PolylinePath):
            entities.append([polyline()])
        elif isinstance(path, EdgePath):
            entities.append(
                _virtual_edge_path(path, dict(graphic_attribs), ocs, elevation)
            )
    return entities


def _virtual_edge_path(
    path: EdgePath, dxfattribs, ocs: OCS, elevation: float
) -> list[DXFGraphic]:
    from ezdxf.entities import Line, Arc, Ellipse, Spline

    def pnt_to_wcs(v):
        return ocs.to_wcs(Vec3(v).replace(z=elevation))

    def dir_to_wcs(v):
        return ocs.to_wcs(v)

    edges: list[DXFGraphic] = []
    for edge in path.edges:
        attribs = dict(dxfattribs)
        if isinstance(edge, LineEdge):
            attribs["start"] = pnt_to_wcs(edge.start)
            attribs["end"] = pnt_to_wcs(edge.end)
            edges.append(Line.new(dxfattribs=attribs))
        elif isinstance(edge, ArcEdge):
            attribs["center"] = edge.center
            attribs["radius"] = edge.radius
            attribs["elevation"] = elevation
            # Arcs angles are always stored in counter-clockwise orientation
            # around the extrusion vector!
            attribs["start_angle"] = edge.start_angle
            attribs["end_angle"] = edge.end_angle
            attribs["extrusion"] = ocs.uz
            edges.append(Arc.new(dxfattribs=attribs))
        elif isinstance(edge, EllipseEdge):
            attribs["center"] = pnt_to_wcs(edge.center)
            attribs["major_axis"] = dir_to_wcs(edge.major_axis)
            attribs["ratio"] = edge.ratio
            # Ellipse angles are always stored in counter-clockwise orientation
            # around the extrusion vector!
            attribs["start_param"] = edge.start_param
            attribs["end_param"] = edge.end_param
            attribs["extrusion"] = ocs.uz
            edges.append(Ellipse.new(dxfattribs=attribs))
        elif isinstance(edge, SplineEdge):
            spline = Spline.new(dxfattribs=attribs)
            spline.dxf.degree = edge.degree
            spline.knots = edge.knot_values
            spline.control_points = [pnt_to_wcs(v) for v in edge.control_points]
            if edge.weights:
                spline.weights = edge.weights
            if edge.fit_points:
                spline.fit_points = [pnt_to_wcs(v) for v in edge.fit_points]
            if edge.start_tangent is not None:
                spline.dxf.start_tangent = dir_to_wcs(edge.start_tangent)
            if edge.end_tangent is not None:
                spline.dxf.end_tangent = dir_to_wcs(edge.end_tangent)
            edges.append(spline)
    return edges



# Copyright 2018, Manfred Moitzi
# License: MIT License
import os
import math
import random
import pytest
import ezdxf
from ezdxf.lldxf.const import versions_supported_by_new

MSIZE = 20
HEIGHT = 3.0


@pytest.fixture(params=versions_supported_by_new)
def drawing(request):
    return ezdxf.new(request.param)


def build_mesh(polymesh):
    m_size = polymesh.dxf.m_count
    n_size = polymesh.dxf.n_count
    m_delta = math.pi / m_size
    n_delta = math.pi / n_size

    for x in range(m_size):
        sinx = math.sin(float(x) * m_delta)
        for y in range(n_size):
            cosy = math.cos(float(y) * n_delta)
            z = sinx * cosy * HEIGHT
            # set the m,n vertex to 3d point x,y,z
            polymesh.set_mesh_vertex(pos=(x, y), point=(x, y, z))


def build_cube(layout, basepoint, length):
    def scale(point):
        return (
            (basepoint[0] + point[0] * length),
            (basepoint[1] + point[1] * length),
            (basepoint[2] + point[2] * length),
        )

    # cube corner points
    p1 = scale((0, 0, 0))
    p2 = scale((0, 0, 1))
    p3 = scale((0, 1, 0))
    p4 = scale((0, 1, 1))
    p5 = scale((1, 0, 0))
    p6 = scale((1, 0, 1))
    p7 = scale((1, 1, 0))
    p8 = scale((1, 1, 1))

    # define the 6 cube faces
    # look into -x direction
    # Every add_face adds 4 vertices 6x4 = 24 vertices
    pface = layout.add_polyface()
    pface.append_face([p1, p5, p7, p3], {"color": 1})  # base
    pface.append_face([p1, p5, p6, p2], {"color": 2})  # left
    pface.append_face([p5, p7, p8, p6], {"color": 3})  # front
    pface.append_face([p7, p8, p4, p3], {"color": 4})  # right
    pface.append_face([p1, p3, p4, p2], {"color": 5})  # back
    pface.append_face([p2, p6, p8, p4], {"color": 6})  # top


def add_polymesh(layout):
    polymesh = layout.add_polymesh(size=(MSIZE, MSIZE))
    build_mesh(polymesh)


def add_polyfaces(layout):
    for x in range(10):
        for y in range(10):
            build_cube(
                layout,
                basepoint=(x, y, random.random()),
                length=random.random(),
            )


def add_polyline2d(layout):
    points2d = [
        (0, 0),
        (1, 0),
        (1, 1),
        (0, 1),
        (0, 0),
        (1, 1),
        (0.5, 1.5),
        (0, 1),
        (1, 0),
    ]
    layout.add_polyline2d(points2d)


def add_polyline3d(layout):
    points3d = [(3, 3, 0), (6, 3, 1), (6, 6, 2), (3, 6, 3), (3, 3, 4)]
    layout.add_polyline3d(points3d)


def test_create_polyline_entities(drawing, tmpdir):
    modelspace = drawing.modelspace()

    add_polyline2d(modelspace)
    add_polyline3d(modelspace)
    add_polyfaces(modelspace)
    add_polymesh(modelspace)

    filename = str(tmpdir.join("polyline_entities_%s.dxf" % drawing.dxfversion))
    try:
        drawing.saveas(filename)
    except ezdxf.DXFError as e:
        pytest.fail(
            "DXFError: {0} for DXF version {1}".format(
                str(e), drawing.dxfversion
            )
        )
    assert os.path.exists(filename)



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
import pytest
import ezdxf
from ezdxf.addons import MTextExplode

LOREM_IPSUM = (
    r"Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam "
    r"nonumy eirmod tempor {\C1invidunt ut labore} et dolore mag{\C3na al}iquyam "
    r"erat, sed {\C5diam voluptua.} At vero eos et accusam et justo duo dolores "
    r"et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est "
    r"Lorem ipsum dolor sit amet."
)
LEFT = LOREM_IPSUM + "\n\n"
CENTER = r"\pxqc;" + LOREM_IPSUM + "\n\n"
RIGHT = r"\pxqr;" + LOREM_IPSUM + "\n\n"
JUSTIFIED = r"\pi1,qj;" + LOREM_IPSUM + "\n\n"


def new_doc(content: str, width: float = 30):
    doc = ezdxf.new(setup=True)
    msp = doc.modelspace()
    mtext = msp.add_mtext(
        content,
        dxfattribs={
            "layer": "MTEXT_EXPLODE",
            "width": width,
            "char_height": 1,
            "color": 7,
            "style": "OpenSans",
            "line_spacing_style": ezdxf.const.MTEXT_EXACT,
        },
    )
    mtext.set_bg_color(None, text_frame=True)
    return doc


def explode_mtext(doc, destroy=False):
    msp = doc.modelspace()
    with MTextExplode(msp) as xpl:
        for mtext in msp.query("MTEXT"):
            xpl.explode(mtext, destroy=destroy)
            if mtext.is_alive:
                mtext.dxf.layer = "SOURCE"


def test_created_text_styles_exists():
    from ezdxf.tools.text import MTextEditor

    doc = ezdxf.new()
    msp = doc.modelspace()
    editor = MTextEditor()
    editor.append("LINE0\n")
    editor.font("Open Sans")
    editor.append("LINE1")
    mtext = msp.add_mtext(editor.text)
    with MTextExplode(msp) as xpl:
        xpl.explode(mtext)
    assert doc.styles.has_entry("MtXpl_DejaVu Sans Condensed")  # default font
    assert doc.styles.has_entry("MtXpl_Open Sans")  # MTEXT inline set font


def test_addon_is_still_working():
    # Testing only the basic functionality!
    # The text layout engine has its own test suite in test file 517!
    # The MTextParser has its own test suite in test file 521!
    doc = new_doc(LEFT + CENTER + RIGHT + JUSTIFIED)
    explode_mtext(doc, destroy=True)
    msp = doc.modelspace()
    assert len(msp.query("MTEXT")) == 0
    assert len(msp.query("TEXT")) == 208  # checked by BricsCAD



#  Copyright (c) 2025, Manfred Moitzi
#  License: MIT License
import pytest

import math
from ezdxf.math import BSpline
from ezdxf.math.bspline import to_homogeneous_points, from_homogeneous_points
import numpy as np

CONTROL_POINTS = [(0, 0), (1, -1), (2, 0), (3, 2), (4, 0), (5, -2)]
WEIGHTS = [1, 2, 3, 3, 2, 1]

# visually checked in BricsCAD
EXPECTED_POINTS = [
    (0.0, 0.0, 0.0),
    (0.75, -0.75, 0.0),
    (1.25, -0.75, 0.0),
    (1.9583333333333335, 0.04166666666666663, 0.0),
    (2.5, 1.0, 0.0),
    (3.041666666666666, 1.583333333333333, 0.0),
    (3.75, 0.5, 0.0),
    (4.25, -0.5, 0.0),
    (5.0, -2.0, 0.0),
]
EXPECTED_KNOTS = [
    0.0,
    0.0,
    0.0,
    0.0,
    0.0,
    0.3333333333333333,
    0.3333333333333333,
    0.6666666666666666,
    0.6666666666666666,
    1.0,
    1.0,
    1.0,
    1.0,
    1.0,
]

EXPECTED_POINTS_RATIONAL = [
    (0.0, 0.0, 0.0),
    (0.8571428571428571, -0.8571428571428571, 0.0),
    (1.3333333333333333, -0.6666666666666666, 0.0),
    (2.0, 0.08695652173913043, 0.0),
    (2.5, 1.0, 0.0),
    (2.999999999999999, 1.6521739130434785, 0.0),
    (3.6666666666666665, 0.6666666666666666, 0.0),
    (4.142857142857143, -0.2857142857142857, 0.0),
    (5.0, -2.0, 0.0),
]

EXPECTED_WEIGHTS = [1.0, 1.75, 2.25, 2.875, 3.0, 2.8749999999999996, 2.25, 1.75, 1.0]


class TestDegreeElevationNonRationalBSpline:
    @pytest.fixture(scope="class")
    def result(self):
        spline = BSpline(CONTROL_POINTS)
        return spline.degree_elevation(1)

    def test_degree_is_elevated(self, result: BSpline):
        assert result.degree == 4

    def test_clamped_start_and_end_points_are_preserved(self, result: BSpline):
        assert result.control_points[0].isclose(CONTROL_POINTS[0])
        assert result.control_points[-1].isclose(CONTROL_POINTS[-1])

    def test_expected_control_points(self, result: BSpline):
        assert (
            all(cp.isclose(e) for cp, e in zip(result.control_points, EXPECTED_POINTS))
            is True
        )

    def test_expected_knot_values(self, result: BSpline):
        assert (
            all(math.isclose(k, e) for k, e in zip(result.knots(), EXPECTED_KNOTS))
            is True
        )

    def test_elevation_0_times(self):
        spline = BSpline(CONTROL_POINTS)
        assert spline.degree_elevation(0) is spline
        assert spline.degree_elevation(-1) is spline


class TestDegreeElevationRationalBSpline:
    @pytest.fixture(scope="class")
    def result(self):
        spline = BSpline(CONTROL_POINTS, weights=WEIGHTS)
        return spline.degree_elevation(1)

    def test_degree_is_elevated(self, result: BSpline):
        assert result.degree == 4

    def test_clamped_start_and_end_points_are_preserved(self, result: BSpline):
        assert result.control_points[0].isclose(CONTROL_POINTS[0])
        assert result.control_points[-1].isclose(CONTROL_POINTS[-1])

    def test_expected_control_points(self, result: BSpline):
        assert (
            all(
                cp.isclose(e)
                for cp, e in zip(result.control_points, EXPECTED_POINTS_RATIONAL)
            )
            is True
        )

    def test_expected_knot_values(self, result: BSpline):
        assert (
            all(math.isclose(k, e) for k, e in zip(result.knots(), EXPECTED_KNOTS))
            is True
        )

    def test_expected_weights(self, result: BSpline):
        assert (
            all(math.isclose(w, e) for w, e in zip(result.weights(), EXPECTED_WEIGHTS))
            is True
        )


class TestHomogeneousPoints:
    def test_to_hg_points(self):
        bsp = BSpline(CONTROL_POINTS, weights=WEIGHTS)
        hp = to_homogeneous_points(bsp)
        assert len(hp) == len(CONTROL_POINTS)
        assert len(hp[0]) == 4
        assert tuple(hp[1]) == (2, -2, 0, 2)

    def test_to_hg_points_non_rational(self):
        bsp = BSpline(CONTROL_POINTS)
        hp = to_homogeneous_points(bsp)
        assert len(hp) == len(CONTROL_POINTS)
        assert len(hp[0]) == 4
        assert tuple(hp[1]) == (1, -1, 0, 1)

    def test_revert_hg_points(self):
        bsp = BSpline(CONTROL_POINTS, weights=WEIGHTS)
        hp = to_homogeneous_points(bsp)
        points, weights = from_homogeneous_points(hp)
        assert len(points) == len(CONTROL_POINTS)
        assert len(weights) == len(WEIGHTS)
        assert all(v.isclose(e) for v, e in zip(points, CONTROL_POINTS)) is True
        assert all(math.isclose(w, e) for w, e in zip(weights, WEIGHTS)) is True


class TestPointInversion:
    @pytest.fixture(scope="class")
    def spline(self):
        return BSpline(CONTROL_POINTS)

    def test_start_point(self, spline):
        assert math.isclose(0, spline.point_inversion((0, 0)))

    def test_end_point(self, spline):
        result = spline.point_inversion(CONTROL_POINTS[-1])
        assert math.isclose(spline.max_t, result)

    def test_many_points(self, spline):
        t = np.linspace(0, spline.max_t, 13)
        points = spline.points(t)
        assert all(
            math.isclose(spline.point_inversion(p), u) for p, u in zip(points, t)
        )


class TestSplineMeasurement:
    # measurements done in BricsCAD v25
    total_length = 7.7842
    mid_length = total_length / 2
    mid_param = 0.6251460418169124

    @pytest.fixture(scope="class")
    def spline(self):
        return BSpline(CONTROL_POINTS)

    def test_measure_length(self, spline):
        mtool = spline.measure(100)
        # measurement diverges less than 1%
        assert mtool.length / self.total_length > 0.99

    def test_measure_length_by_0_segments(self, spline):
        with pytest.raises(ValueError):
            spline.measure(0)

    def test_get_param_for_mid_point(self, spline):
        mtool = spline.measure(100)
        t = mtool.param_at(self.mid_length)
        # measurement diverges less than 1%
        assert 0.99 < t / self.mid_param <= 1.0

    def test_params_out_of_range(self, spline):
        mtool = spline.measure(100)
        assert mtool.param_at(-1) == 0.0
        assert mtool.param_at(100) == spline.max_t

    def test_distance_from_start(self, spline):
        mtool = spline.measure(100)
        ratio = mtool.distance(self.mid_param) / self.mid_length
        # measurement diverges less than 1%
        assert 0.99 < ratio <= 1.00

    def test_distance_out_of_range(self, spline):
        mtool = spline.measure(100)
        assert mtool.distance(-1) == 0.0
        assert mtool.distance(100) == mtool.length

    def test_divide(self, spline):
        mtool = spline.measure(100)
        params = mtool.divide(7)
        assert len(params) == 6
        assert params[0] != 0.0
        assert params[-1] != spline.max_t

    def test_extents(self, spline):
        mtool = spline.measure()
        assert mtool.extmin.isclose((0, -2))
        assert mtool.extmax.isclose((5, 1.19263675))


def test_split_spline():
    spline = BSpline(CONTROL_POINTS)
    mid_t = spline.measure().divide(2)[0]
    sp1, sp2 = spline.split(mid_t)
    l1 = sp1.measure().length
    l2 = sp2.measure().length
    assert abs(l1 - l2) < 0.01


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2013-2026, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    Iterable,
    Iterator,
    Callable,
    Hashable,
    Sequence,
    Union,
    Optional,
)
import re
import operator
from collections import abc

from ezdxf.entities.dxfentity import DXFEntity
from ezdxf.groupby import groupby
from ezdxf.math import Vec3, Vec2
from ezdxf.queryparser import EntityQueryParser


class _AttributeDescriptor:
    def __init__(self, name: str):
        self.name = name

    def __get__(self, obj, objtype=None):
        return obj.__getitem__(self.name)

    def __set__(self, obj, value):
        obj.__setitem__(self.name, value)

    def __delete__(self, obj):
        obj.__delitem__(self.name)


class EntityQuery(abc.Sequence):
    """EntityQuery is a result container, which is filled with dxf entities
    matching the query string. It is possible to add entities to the container
    (extend), remove entities from the container and to filter the container.

    Query String
    ============

    QueryString := EntityQuery ("[" AttribQuery "]")*

    The query string is the combination of two queries, first the required
    entity query and second the optional attribute query, enclosed in square
    brackets.

    Entity Query
    ------------

    The entity query is a whitespace separated list of DXF entity names or the
    special name ``*``. Where ``*`` means all DXF entities, exclude some entity
    types by appending their names with a preceding ``!`` (e.g. all entities
    except LINE = ``* !LINE``). All DXF names have to be uppercase.

    Attribute Query
    ---------------

    The attribute query is used to select DXF entities by its DXF attributes.
    The attribute query is an addition to the entity query and matches only if
    the entity already match the entity query.
    The attribute query is a boolean expression, supported operators are:

      - not: !term is true, if term is false
      - and: term & term is true, if both terms are true
      - or: term | term is true, if one term is true
      - and arbitrary nested round brackets

    Attribute selection is a term: "name comparator value", where name is a DXF
    entity attribute in lowercase, value is a integer, float or double quoted
    string, valid comparators are:

      - "==" equal "value"
      - "!=" not equal "value"
      - "<" lower than "value"
      - "<=" lower or equal than "value"
      - ">" greater than "value"
      - ">=" greater or equal than "value"
      - "?" match regular expression "value"
      - "!?" does not match regular expression "value"

    Query Result
    ------------

    The EntityQuery() class based on the abstract Sequence() class, contains all
    DXF entities of the source collection, which matches one name of the entity
    query AND the whole attribute query. If a DXF entity does not have or
    support a required attribute, the corresponding attribute search term is
    false.

    Examples:

        - 'LINE[text ? ".*"]' is always empty, because the LINE entity has no
          text attribute.
        - 'LINE CIRCLE[layer=="construction"]' => all LINE and CIRCLE entities
          on layer "construction"
        - '*[!(layer=="construction" & color<7)]' => all entities except those
          on layer == "construction" and color < 7

    """

    layer = _AttributeDescriptor("layer")
    color = _AttributeDescriptor("color")
    linetype = _AttributeDescriptor("linetype")
    lineweight = _AttributeDescriptor("lineweight")
    ltscale = _AttributeDescriptor("ltscale")
    invisible = _AttributeDescriptor("invisible")
    true_color = _AttributeDescriptor("true_color")
    transparency = _AttributeDescriptor("transparency")

    def __init__(
        self, entities: Optional[Iterable[DXFEntity]] = None, query: str = "*"
    ):
        """
        Setup container with entities matching the initial query.

        Args:
            entities: sequence of wrapped DXF entities (at least GraphicEntity class)
            query: query string, see class documentation

        """
        # Selected DXF attribute for operator selection:
        self.selected_dxf_attribute: str = ""
        # Text selection mode, but only for operator comparisons:
        self.ignore_case = True

        self.entities: list[DXFEntity]
        if entities is None:
            self.entities = []
        elif query == "*":
            self.entities = list(entities)
        else:
            match = entity_matcher(query)
            self.entities = [entity for entity in entities if match(entity)]

    def __len__(self) -> int:
        """Returns count of DXF entities."""
        return len(self.entities)

    def __iter__(self) -> Iterator[DXFEntity]:
        """Returns iterable of DXFEntity objects."""
        return iter(self.entities)

    def __getitem__(self, item):
        """Returns DXFEntity at index `item`, supports negative indices and
        slicing. Returns all entities which support a specific DXF attribute,
        if `item` is a DXF attribute name as string.
        """
        if isinstance(item, str):
            return self._get_entities_with_supported_attribute(item)
        return self.entities.__getitem__(item)

    def __setitem__(self, key, value):
        """Set the DXF attribute `key` for all supported DXF entities to `value`.
        """
        if not isinstance(key, str):
            raise TypeError("key has to be a string (DXF attribute name)")
        self._set_dxf_attribute_for_all(key, value)

    def __delitem__(self, key):
        """Discard the DXF attribute `key` from all supported DXF entities."""
        if not isinstance(key, str):
            raise TypeError("key has to be a string (DXF attribute name)")
        self._discard_dxf_attribute_for_all(key)

    def purge(self) -> EntityQuery:
        """Remove destroyed entities."""
        self.entities = [e for e in self.entities if e.is_alive]
        return self  # fluent interface

    def _get_entities_with_supported_attribute(
        self, attribute: str
    ) -> EntityQuery:
        query = self.__class__(
            e for e in self.entities if e.dxf.is_supported(attribute)
        )
        query.selected_dxf_attribute = attribute
        return query

    def _set_dxf_attribute_for_all(self, key, value):
        for e in self.entities:
            try:
                e.dxf.set(key, value)
            except AttributeError:  # ignore unsupported attributes
                pass
            # But raises ValueError/TypeError for invalid values!

    def _discard_dxf_attribute_for_all(self, key):
        for e in self.entities:
            e.dxf.discard(key)

    def __eq__(self, other):
        """Equal selector (self == other).
        Returns all entities where the selected DXF attribute is equal to
        `other`.
        """
        if not self.selected_dxf_attribute:
            raise TypeError("no DXF attribute selected")
        return self._select_by_operator(other, operator.eq)

    def __ne__(self, other):
        """Not equal selector (self != other). Returns all entities where the
        selected DXF attribute is not equal to `other`.
        """
        if not self.selected_dxf_attribute:
            raise TypeError("no DXF attribute selected")
        return self._select_by_operator(other, operator.ne)

    def __lt__(self, other):
        """Less than selector (self < other). Returns all entities where the
        selected DXF attribute is less than `other`.

        Raises:
             TypeError: for vector based attributes like `center` or `insert`
        """
        if not self.selected_dxf_attribute:
            raise TypeError("no DXF attribute selected")
        return self._select_by_operator(other, operator.lt, vectors=False)

    def __gt__(self, other):
        """Greater than selector (self > other). Returns all entities where the
        selected DXF attribute is greater than `other`.

        Raises:
             TypeError: for vector based attributes like `center` or `insert`
        """
        if not self.selected_dxf_attribute:
            raise TypeError("no DXF attribute selected")
        return self._select_by_operator(other, operator.gt, vectors=False)

    def __le__(self, other):
        """Less equal selector (self <= other). Returns all entities where the
        selected DXF attribute is less or equal `other`.

        Raises:
             TypeError: for vector based attributes like `center` or `insert`
        """
        if not self.selected_dxf_attribute:
            raise TypeError("no DXF attribute selected")
        return self._select_by_operator(other, operator.le, vectors=False)

    def __ge__(self, other):
        """Greater equal selector (self >= other). Returns all entities where
        the selected DXF attribute is greater or equal `other`.

        Raises:
             TypeError: for vector based attributes like `center` or `insert`
        """
        if not self.selected_dxf_attribute:
            raise TypeError("no DXF attribute selected")
        return self._select_by_operator(other, operator.ge, vectors=False)

    def __or__(self, other):
        """Union operator, see :meth:`union`."""
        if isinstance(other, EntityQuery):
            return self.union(other)
        return NotImplemented

    def __and__(self, other):
        """Intersection operator, see :meth:`intersection`."""
        if isinstance(other, EntityQuery):
            return self.intersection(other)
        return NotImplemented

    def __sub__(self, other):
        """Difference operator, see :meth:`difference`."""
        if isinstance(other, EntityQuery):
            return self.difference(other)
        return NotImplemented

    def __xor__(self, other):
        """Symmetric difference operator, see :meth:`symmetric_difference`."""
        if isinstance(other, EntityQuery):
            return self.symmetric_difference(other)
        return NotImplemented

    def _select_by_operator(self, value, op, vectors=True) -> EntityQuery:
        attribute = self.selected_dxf_attribute
        if self.ignore_case and isinstance(value, str):
            value = value.lower()

        query = self.__class__()
        query.selected_dxf_attribute = attribute
        entities = query.entities
        if attribute:
            for entity in self.entities:
                try:
                    entity_value = entity.dxf.get_default(attribute)
                except AttributeError:
                    continue
                if not vectors and isinstance(entity_value, (Vec2, Vec3)):
                    raise TypeError(
                        f"unsupported operation '{str(op.__name__)}' for DXF "
                        f"attribute {attribute}"
                    )
                if self.ignore_case and isinstance(entity_value, str):
                    entity_value = entity_value.lower()
                if op(entity_value, value):
                    entities.append(entity)
        return query

    def match(self, pattern: str) -> EntityQuery:
        """Returns all entities where the selected DXF attribute matches the
        regular expression `pattern`.

        Raises:
             TypeError: for non-string based attributes

        """

        def match(value, regex):
            if isinstance(value, str):
                return regex.match(value) is not None
            raise TypeError(
                f"cannot apply regular expression to DXF attribute: "
                f"{self.selected_dxf_attribute}"
            )

        return self._regex_match(pattern, match)

    def _regex_match(self, pattern: str, func) -> EntityQuery:
        ignore_case = self.ignore_case
        self.ignore_case = False  # deactivate string manipulation
        re_flags = re.IGNORECASE if ignore_case else 0

        # always match whole pattern
        if not pattern.endswith("$"):
            pattern += "$"
        result = self._select_by_operator(
            re.compile(pattern, flags=re_flags), func
        )
        self.ignore_case = ignore_case  # restore state
        return result

    @property
    def first(self):
        """First entity or ``None``."""
        if len(self.entities):
            return self.entities[0]
        else:
            return None

    @property
    def last(self):
        """Last entity or ``None``."""
        if len(self.entities):
            return self.entities[-1]
        else:
            return None

    def extend(
        self,
        entities: Iterable[DXFEntity],
        query: str = "*",
    ) -> EntityQuery:
        """Extent the :class:`EntityQuery` container by entities matching an
        additional query.

        """
        self.entities = self.union(self.__class__(entities, query)).entities
        return self  # fluent interface

    def remove(self, query: str = "*") -> EntityQuery:
        """Remove all entities from :class:`EntityQuery` container matching this
        additional query.

        """
        self.entities = self.difference(
            self.__class__(self.entities, query)
        ).entities
        return self  # fluent interface

    def query(self, query: str = "*") -> EntityQuery:
        """Returns a new :class:`EntityQuery` container with all entities
        matching this additional query.

        Raises:
            pyparsing.ParseException: query string parsing error

        """
        return self.__class__(self.entities, query)

    def groupby(
        self,
        dxfattrib: str = "",
        key: Optional[Callable[[DXFEntity], Hashable]] = None,
    ) -> dict[Hashable, list[DXFEntity]]:
        """Returns a dict of entity lists, where entities are grouped by a DXF
        attribute or a key function.

        Args:
            dxfattrib: grouping DXF attribute as string like ``'layer'``
            key: key function, which accepts a DXFEntity as argument, returns
                grouping key of this entity or ``None`` for ignore this object.
                Reason for ignoring: a queried DXF attribute is not supported by
                this entity

        """
        return groupby(self.entities, dxfattrib, key)

    def filter(self, func: Callable[[DXFEntity], bool]) -> EntityQuery:
        """Returns a new :class:`EntityQuery` with all entities from this
        container for which the callable `func` returns ``True``.

        Build your own operator to filter by attributes which are not DXF
        attributes or to build complex queries::

            result = msp.query().filter(
                lambda e: hasattr(e, "rgb") and e.rbg == (0, 0, 0)
            )
        """
        return self.__class__(filter(func, self.entities))

    def union(self, other: EntityQuery) -> EntityQuery:
        """Returns a new :class:`EntityQuery` with entities from `self` and
        `other`. All entities are unique - no duplicates.
        """
        return self.__class__(set(self.entities) | set(other.entities))

    def intersection(self, other: EntityQuery) -> EntityQuery:
        """Returns a new :class:`EntityQuery` with entities common to `self`
        and `other`.
        """
        return self.__class__(set(self.entities) & set(other.entities))

    def difference(self, other: EntityQuery) -> EntityQuery:
        """Returns a new :class:`EntityQuery` with all entities from `self` that
        are not in `other`.
        """
        return self.__class__(set(self.entities) - set(other.entities))

    def symmetric_difference(self, other: EntityQuery) -> EntityQuery:
        """Returns a new :class:`EntityQuery` with entities in either `self` or
        `other` but not both.
        """
        return self.__class__(set(self.entities) ^ set(other.entities))


def entity_matcher(query: str) -> Callable[[DXFEntity], bool]:
    query_args = EntityQueryParser.parse_string(query, parse_all=True)
    entity_matcher_ = build_entity_name_matcher(query_args.EntityQuery)
    attrib_matcher = build_entity_attributes_matcher(
        query_args.AttribQuery, query_args.AttribQueryOptions
    )

    def matcher(entity: DXFEntity) -> bool:
        return entity_matcher_(entity) and attrib_matcher(entity)

    return matcher


def build_entity_name_matcher(
    names: Sequence[str],
) -> Callable[[DXFEntity], bool]:
    def match(e: DXFEntity) -> bool:
        return _match(e.dxftype())

    _match = name_matcher(query=" ".join(names))
    return match


class Relation:
    CMP_OPERATORS = {
        "==": operator.eq,
        "!=": operator.ne,
        "<": operator.lt,
        "<=": operator.le,
        ">": operator.gt,
        ">=": operator.ge,
        "?": lambda e, regex: regex.match(e) is not None,
        "!?": lambda e, regex: regex.match(e) is None,
    }
    VALID_CMP_OPERATORS = frozenset(CMP_OPERATORS.keys())

    def __init__(self, relation: Sequence, ignore_case: bool):
        name, op, value = relation
        self.dxf_attrib = name
        self.compare = Relation.CMP_OPERATORS[op]
        self.convert_case = to_lower if ignore_case else lambda x: x

        re_flags = re.IGNORECASE if ignore_case else 0
        if "?" in op:
            self.value = re.compile(
                value + "$", flags=re_flags
            )  # always match whole pattern
        else:
            self.value = self.convert_case(value)

    def evaluate(self, entity: DXFEntity) -> bool:
        try:
            value = self.convert_case(entity.dxf.get_default(self.dxf_attrib))
            return self.compare(value, self.value)
        except AttributeError:  # entity does not support this attribute
            return False
        except ValueError:  # entity supports this attribute, but has no value for it
            return False


def to_lower(value):
    return value.lower() if hasattr(value, "lower") else value


class BoolExpression:
    OPERATORS = {
        "&": operator.and_,
        "|": operator.or_,
    }

    def __init__(self, tokens: Sequence):
        self.tokens = tokens

    def __iter__(self):
        return iter(self.tokens)

    def evaluate(self, entity: DXFEntity) -> bool:
        if isinstance(
            self.tokens, Relation
        ):  # expression is just one relation, no bool operations
            return self.tokens.evaluate(entity)

        values = []  # first in, first out
        operators = []  # first in, first out
        for token in self.tokens:
            if hasattr(token, "evaluate"):
                values.append(token.evaluate(entity))
            else:  # bool operator
                operators.append(token)
        values.reverse()
        for op in operators:  # as queue -> first in, first out
            if op == "!":
                value = not values.pop()
            else:
                value = BoolExpression.OPERATORS[op](values.pop(), values.pop())
            values.append(value)
        return values.pop()


def _compile_tokens(
    tokens: Union[str, Sequence], ignore_case: bool
) -> Union[str, Relation, BoolExpression]:
    def is_relation(tokens: Sequence) -> bool:
        return len(tokens) == 3 and tokens[1] in Relation.VALID_CMP_OPERATORS

    if isinstance(tokens, str):  # bool operator as string
        return tokens

    tokens = tuple(tokens)
    if is_relation(tokens):
        return Relation(tokens, ignore_case)
    else:
        return BoolExpression(
            [_compile_tokens(token, ignore_case) for token in tokens]
        )


def build_entity_attributes_matcher(
    tokens: Sequence, options: str
) -> Callable[[DXFEntity], bool]:
    if not len(tokens):
        return lambda x: True
    ignore_case = "i" == options  # at this time just one option is supported
    expr = BoolExpression(_compile_tokens(tokens, ignore_case))  # type: ignore

    def match_bool_expr(entity: DXFEntity) -> bool:
        return expr.evaluate(entity)

    return match_bool_expr


def unique_entities(entities: Iterable[DXFEntity]) -> Iterator[DXFEntity]:
    """Yield all unique entities, order of all entities will be preserved."""
    done: set[DXFEntity] = set()
    for entity in entities:
        if entity not in done:
            done.add(entity)
            yield entity


def name_query(names: Iterable[str], query: str = "*") -> Iterator[str]:
    """Filters `names` by `query` string. The `query` string of entity names
    divided by spaces. The special name "*" matches any given name, a
    preceding "!" means exclude this name. Excluding names is only useful if
    the match any name is also given (e.g. "LINE !CIRCLE" is equal to just
    "LINE", where "* !CIRCLE" matches everything except CIRCLE").

    Args:
        names: iterable of names to test
        query: query string of entity names separated by spaces

    Returns: yield matching names

    """
    match = name_matcher(query)
    return (name for name in names if match(name))


def name_matcher(query: str = "*") -> Callable[[str], bool]:
    def match(e: str) -> bool:
        if take_all:
            return e not in exclude
        else:
            return e in include

    match_strings = set(query.upper().split())
    take_all = False
    exclude = set()
    include = set()
    for name in match_strings:
        if name == "*":
            take_all = True
        elif name.startswith("!"):
            exclude.add(name[1:])
        else:
            include.add(name)

    return match


def new(
    entities: Optional[Iterable[DXFEntity]] = None, query: str = "*"
) -> EntityQuery:
    """Start a new query based on sequence `entities`. The `entities` argument
    has to be an iterable of :class:`~ezdxf.entities.DXFEntity` or inherited
    objects and returns an :class:`EntityQuery` object.

    """
    return EntityQuery(entities, query)



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
# created 2019-02-15
import pytest

from ezdxf.entities.underlay import PdfDefinition, DgnDefinition, DwfDefinition
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

PDFDEFINITION = """0
PDFDEFINITION
5
0
330
0
100
AcDbUnderlayDefinition
1
doc.pdf
2
1
"""


@pytest.fixture
def entity():
    return PdfDefinition.from_text(PDFDEFINITION)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "PDFDEFINITION" in ENTITY_CLASSES


def test_default_init():
    entity = PdfDefinition()
    assert entity.dxftype() == "PDFDEFINITION"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = PdfDefinition.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "filename": "doc.pdf",
            "name": "1",
        },
    )
    assert entity.dxf.filename == "doc.pdf"
    assert entity.dxf.name == "1"


def test_load_from_text(entity):
    assert entity.dxf.filename == "doc.pdf"
    assert entity.dxf.name == "1"


def test_write_dxf():
    entity = PdfDefinition.from_text(PDFDEFINITION)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(PDFDEFINITION)
    assert result == expected


def test_dwf_definition():
    assert DwfDefinition.DXFTYPE == "DWFDEFINITION"


def test_dgn_definition():
    assert DgnDefinition.DXFTYPE == "DGNDEFINITION"
